﻿using System;
using System.DirectoryServices;
using System.DirectoryServices.AccountManagement;
using System.Globalization;
using System.Runtime.InteropServices;
using System.Security.Principal;

namespace com.swivelsecure.pinsafe {
	public class ActiveDirectoryClient {

		public const int ERR_OK = 0;
		public const int ERR_PWD_CHANGE_REQ = 1907;
		public const int ERR_PWD_EXPIRED = 1330;
		public const int ERR_LOGON_TYPE_NOT_ALLOWED = 1385;
		public const int ERR_LOGON_FAILED = 1326;
		public const int ERR_AD_DISABLED = -1;

		private const int LOGON_INTERACTIVE = 2;
		private const int LOGON_NETWORK = 3;
		private const int LOGON_BATCH = 4;
		private static readonly int[] LOGON_TYPES = new[] { LOGON_INTERACTIVE, LOGON_NETWORK, LOGON_BATCH };

		private const string LOG_SOURCE = "AD_CLIENT";

        private readonly PINsafeSettings settings;
        private readonly TextSettings text;

		private bool Enabled {
			get { return settings.ManageADPasswords; }
		}

        [DllImport("ADVAPI32.dll", EntryPoint = "LogonUserW", SetLastError = true, CharSet = CharSet.Auto)]
        private static extern bool LogonUser(string username, string domain, string password, int logonType,
                                            int logonProvider, ref IntPtr token);

        private static string GetDomainName(string usernameDomain) {
            if (string.IsNullOrEmpty(usernameDomain)) {
                throw (new ArgumentException("Argument can't be null.", "usernameDomain"));
            }
            if (usernameDomain.Contains("\\")) {
                int index = usernameDomain.IndexOf("\\", StringComparison.Ordinal);
                return usernameDomain.Substring(0, index).ToUpper(CultureInfo.InvariantCulture);
            }
            if (usernameDomain.Contains("@")) {
                // If this is a userPrincipalName, use it as is, with null domain
                return null;
            }
            return "";
        }

        private static string GetUsername(string usernameDomain) {
            if (string.IsNullOrEmpty(usernameDomain)) {
                throw (new ArgumentException("Argument can't be null.", "usernameDomain"));
            }
            if (usernameDomain.Contains("\\")) {
                int index = usernameDomain.IndexOf("\\", StringComparison.Ordinal);
                return usernameDomain.Substring(index + 1);
            }
            return usernameDomain;
        }  

        public ActiveDirectoryClient(PINsafeSettings _settings, TextSettings _text) {
            settings = _settings;
            text = _text;
        }

        public string Error { get; private set; }

        public ActiveDirectoryClient() {
            settings = PINsafeSettings.SettingsFromConfiguration;
            text = new TextSettings();
        }

		private int LogonUser(string fullUsername, string password, ref IntPtr token) {
			if (!Enabled) {
				Error = text.ChangePasswordDisabled;
				return ERR_AD_DISABLED;
			}
			string username = GetUsername(fullUsername);
			string domain = GetDomainName(fullUsername);
			int typeIdx = 0;
			int retval = 0;
			while (typeIdx < LOGON_TYPES.Length) {
				retval = LogonUser(username, domain, password, LOGON_TYPES[typeIdx], 0, ref token) ? 0 : Marshal.GetLastWin32Error();
				bool success = (token != IntPtr.Zero);
				settings.Logger.WriteLine(LOG_SOURCE, "LogonUser with username=\"{0}\", domain=\"{1}\", type {2} returned error code {3} - {4}", username, domain, LOGON_TYPES[typeIdx], retval, success ? "succeeded" : "failed");
				if (retval == ERR_LOGON_TYPE_NOT_ALLOWED) {
					typeIdx++;
					continue;
				}
				break;
			}
			return retval;
		}

		public int CheckPassword(string fullUsername, string password) {
			if (!Enabled) {
				Error = text.ChangePasswordDisabled;
				return ERR_AD_DISABLED;
			}
			IntPtr token = IntPtr.Zero;
			try {
				return LogonUser(fullUsername, password, ref token);
			} finally {
				if (token != IntPtr.Zero) {
					Marshal.FreeCoTaskMem(token);
				}
			}
		}

		public bool ChangePassword(string username, string oldPassword, string newPassword) {
			if (!Enabled) {
				Error = text.ChangePasswordDisabled;
				return false;
			}
			try {
                using (PrincipalContext context = GetContext()) {
                    if (context == null) {
                        return false;
                    }
                    using (UserPrincipal user = FindUser(context, username)) {
                        if (user == null) {
                            Error = String.Format(text.UserNotFound, username);
                            return false;
                        }
                        user.ChangePassword(oldPassword, newPassword);
                        user.Save();
                    }
                }
            } catch (PasswordException e) {
                if (e.Message.IndexOf("HRESULT: 0x80070056", StringComparison.OrdinalIgnoreCase) > 0) {
                    Error = text.IncorrectOldPassword;
                } else if (e.Message.IndexOf("HRESULT: 0x800708C5", StringComparison.OrdinalIgnoreCase) > 0) {
                    Error = text.PasswordPolicyError;
                } else {
                    Error = text.ChangePasswordError;
                }
                return false;
            }
            return true;
        }

		public string ResetPassword(string username) {
			if (!Enabled) {
				Error = text.ChangePasswordDisabled;
				return null;
			}
			string password = PasswordGenerator.Instance.GenerateRandomPassword();
			return ResetPassword(username, password) ? password : null;
		}

		public bool ResetPassword(string username, string newPassword) {
			bool result = false;
			try {
				using (PrincipalContext context = GetContext()) {
					if (context != null) {
						using (UserPrincipal user = FindUser(context, username)) {
							if (user == null) {
								Error = String.Format(text.UserNotFound, username);
							} else {
								IntPtr token = IntPtr.Zero;
								try {
									if (LogonUser(settings.UserAccount, settings.UserPassword, ref token) == 0) {
										using (WindowsIdentity identity = new WindowsIdentity(token)) {
											WindowsImpersonationContext winContext = identity.Impersonate();
											user.SetPassword(newPassword);
											user.Save();
											winContext.Undo();
											result = true;
										}
									} else {
										Error = "Invalid credentials for Active Directory";
									}
								} finally {
									if (token != IntPtr.Zero) {
										Marshal.FreeCoTaskMem(token);
									}
								}
							}
						}
					}
				}
			} catch (Exception e) {
				if (e.InnerException != null) {
					e = e.InnerException;
				}
				Error = e.Message.IndexOf("HRESULT: 0x800708C5", StringComparison.OrdinalIgnoreCase) > 0 
					? text.PasswordPolicyError
					: e.Message.IndexOf("HRESULT: 0x80070005", StringComparison.OrdinalIgnoreCase) > 0
					? "The specified AD user does not have permission to reset passwords"
					: e.Message;
			}
			if (result) {
				settings.Logger.WriteLine(LOG_SOURCE, text.ResetPasswordSuccess, username);
				settings.Logger.WriteLine(LOG_SOURCE, "New Password is \"{0}\"", newPassword);
			} else {
				settings.Logger.WriteLine(LOG_SOURCE, text.ResetPasswordFailed, username, (String.IsNullOrEmpty(Error)) ? "" : Error);
			}
			return result;
		}

		public bool SendResetPasswordAlert(string username, string password) {
			string message = string.Format(text.ResetPasswordAlert, password);
			PINsafeClient pinsafe = new PINsafeClient(settings, text);
			return pinsafe.SendCustomMessage(username, message);
		}

        public bool PasswordChangeRequired(string username) {
            using (PrincipalContext context = GetContext()) {
                if (context == null) {
                    return false;
                }
	            try {
		            using (UserPrincipal user = FindUser(context, username)) {
			            if (user == null) {
				            Error = "User \"" + username + "\" not found";
				            return false;
			            }
			            // Research shows that this is the way to do it.
			            DateTime? lastPasswordChange = user.LastPasswordSet;
			            if (lastPasswordChange == null) {
				            return (!user.PasswordNeverExpires);
			            }
			            DirectoryEntry entry = user.GetUnderlyingObject() as DirectoryEntry;
			            if (entry != null) {
				            try {
					            entry.RefreshCache(new[] {"msDS-User-Account-Control-Computed"});
					            PropertyValueCollection accountFlags =
						            entry.Properties["msDS-User-Account-Control-Computed"];
					            if (accountFlags == null || accountFlags.Count == 0) {
						            Error = "UAC flags are empty";
					            } else {
						            object oFlags = accountFlags.Value;
						            if (oFlags == null) {
							            Error = "UAC flags are empty";
						            } else {
							            if (!(oFlags is int)) {
								            Error = "Unexpected flags type: " + oFlags.GetType();
							            } else {
								            int flags = (int) oFlags;
								            bool expired = ((flags & 0x800000) != 0);
								            return expired;
							            }
						            }
					            }
				            } catch (Exception e) {
					            Error = "Failed to get UAC flags: " + e.Message;
				            }
			            } else {
				            Error = "Unable to get DirectoryEntry";
			            }
		            }
	            } catch {
		            // If the user doesn't exist, this happens.
//					settings.Logger.WriteLine(LOG_SOURCE, "An error occurred checking user \"{0}\": {1}", username, e.Message);
					Error = "User \"" + username + "\" not found";
					return false;
	            }
            }
            return false;
        }

        public bool UserIsMemberOf(string username, string groupName) {
            using (PrincipalContext context = GetContext()) {
                if (context == null) {
                    return false;
                }
                using (UserPrincipal user = FindUser(context, username)) {
                    if (user == null) {
                        Error = "User \"" + username + "\" not found";
                        return false;
                    }
                    using (GroupPrincipal group = FindGroup(context, groupName)) {
                        if (group == null) {
                            Error = "Group \"" + groupName + "\" not found";
                            return false;
                        }
                        return user.IsMemberOf(group);
                    }
                }
            }
        }

        private PrincipalContext GetContext() {
            PrincipalContext context = new PrincipalContext(ContextType.Domain, null, settings.UserAccount, settings.UserPassword);
            try {
                context.ValidateCredentials(settings.UserAccount, settings.UserPassword);
                return context;
            } catch {
                // Credentials are not valid
                Error = "Active Directory settings invalid";
                return null;
            }
        }

        private UserPrincipal FindUser(PrincipalContext context, string username) {
            IdentityType nametype = (settings.UsernameType == PINsafeSettings.UsernameTypes.userPrincipalName) ? IdentityType.UserPrincipalName : IdentityType.SamAccountName;
            return UserPrincipal.FindByIdentity(context, nametype, username);
        }

        private GroupPrincipal FindGroup(PrincipalContext context, string groupName) {
            return GroupPrincipal.FindByIdentity(context, groupName);
        }
    }
}
