using System;
using System.Drawing;
using System.Globalization;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Web;
using System.Xml;

namespace com.swivelsecure.pinsafe {

    public enum CookieType {
        PINsafe,
        Single,
        ChangePIN
    }

    /// <summary><c>PINsafeClient</c> provides conventient access to the
    ///		authentication service of a PINsafe server</summary>
    public class PINsafeClient {

	    private const string LOG_SOURCE = "PINSAFE_CLIENT";
	    private const string AGENT_XML = "AgentXML";
	    private const string SAS_REQUEST = "SASRequest";
		private const string ACTION = "Action";
		private const string USERNAME = "Username";
		private const string SECRET = "Secret";
		private const string SESSION_ID = "SessionID";
	    private const string PASSWORD = "Password";
	    private const string OTC = "OTC";
	    private const string NEW_PASSWORD = "NewPassword";
	    private const string NEW_OTC = "NewOTC";
	    private const string RESET_CODE = "ResetCode";
	    private const string PASS = "PASS";
	    private const string WARNING = "Warning";
	    private const string ERROR = "Error";
	    private const string RESULT = "Result";

		private const string ADMIN_XML = "AdminXML";
	    //private const string ADMIN_REQUEST = "AdminRequest";
		private const string HELPDESK_REQUEST = "HelpdeskRequest";
		private const string SECRET_ATTR = "secret";
	    private const string VERSION_ATTR = "version";
	    private const string ADMIN_VERSION = "3.96";
	    private const string ADMIN_MESSAGE = "Message";
	    private const string ADMIN_USER = "User";
	    private const string NAME_ATTR = "name";
	    private const string ADMIN_ALERT = "Alert";
		private const string TEXT_ATTR = "text";
	    private const string REPOS_ATTR = "repository";

		private const string ACTION_PING = "ping";
		private const string ACTION_EXISTS = "exists";
		private const string ACTION_SESSION_START = "sessionstart";
		private const string ACTION_LOGIN = "login";
		private const string ACTION_CHANGE_PIN = "changepin";
		private const string ACTION_RESET_CODE = "resetcode";
		private const string ACTION_RESET = "reset";

	    // Server settings

        private readonly PINsafeSettings settings;
        private readonly TextSettings text;

        // Last request/response
        private XmlDocument responseDoc;

        private DebugLogger Debug {
            get { return settings.Logger; }
        }

        public PINsafeClient() : this(PINsafeSettings.SettingsFromConfiguration, null) {
        }

        public PINsafeClient(PINsafeSettings settings) : this(settings, null) {
        }

        public PINsafeClient(PINsafeSettings settings, TextSettings text) {
            this.settings = settings;
            this.text = text;
        }

        /// <summary>Gets or sets the use of SSL for communication with the
        ///		PINsafe server</summary>
        private bool SSL {
            get {
                return settings.Ssl;
            }
        }

        /// <summary>Gets or sets the hostname or IP address of the PINsafe
        ///		server.</summary>
        private string Server {
            get {
                return settings.Server;
            }
        }

        /// <summary>Gets or sets the port used by the PINsafe server.</summary>
        private int Port {
            get {
                return settings.Port;
            }
        }

        /// <summary>Gets or sets the context used by the PINsafe server.</summary>
        private string Context {
            get {
                return settings.Context;
            }
        }

        /// <summary>Gets or sets the secret used for communicating with the
        ///		PINsafe server.</summary>
        private string Secret {
            get {
                return settings.Secret;
            }
        }

        private bool AcceptSelfSigned {
            get { return settings.AcceptSelfSigned; }
        }

        private string CookieSecret {
            get { return settings.CookieSecret; }
        }

        private bool AllowNonPINsafeUsers {
            get { return settings.AllowNonPINsafeUsers; }
        }

        /// <summary>The URL for accessing the PINsafe server.</summary>
        private string AgentXmlAddress {
			get { return GetAddress(AGENT_XML); }
		}

		/// <summary>The URL for accessing the PINsafe server.</summary>
		private string AdminXmlAddress {
			get { return GetAddress(ADMIN_XML); }
		}

		/// <summary>The URL for accessing the PINsafe server.</summary>
		private string GetAddress(string servlet) {
                StringBuilder addressBuilder = new StringBuilder();

                addressBuilder.AppendFormat("{0}://{1}:{2}/{3}/{4}", Protocol, Server, Port, Context, servlet);

                return addressBuilder.ToString();
        }

	    private string Protocol {
		    get { return SSL ? "https" : "http"; }
	    }

	    /// <summary>The URL for accessing the PINsafe server.</summary>
        private string ImageAddress {
            get {
                return GetServletAddress("SCImage");
            }
        }

        /// <summary>The URL for accessing the PINsafe server.</summary>
        private string GetServletAddress(String servlet) {
            StringBuilder addressBuilder = new StringBuilder();

            addressBuilder.AppendFormat("{0}://{1}:{2}/{3}/{4}", Protocol, Server, Port, Context, servlet);

            return addressBuilder.ToString();
        }

        /// <summary>The URL for accessing the PINsafe server.</summary>
        private string MessageAddress {
            get {
                return GetServletAddress("DCMessage");
            }
        }

        /// <summary>Gets the last error.</summary>
        private string Error { get; set; }

        public string ErrorMessage {
            get { return (text == null) ? Error : text.GetText(Error); }
        }

        /// <summary>Gets the last warnings.</summary>
        public string[] Warnings { get; private set; }

        /// <summary>Gets the last XML request sent to the PINSafe server.</summary>
        public string Request { get; private set; }

        /// <value>Gets the last XML response received from the PINsafe server.</value>
        public string Response { get; private set; }

        public bool Ping() {
            try {
                // Create XML
                StringWriter buffer = new StringWriter();
                XmlTextWriter writer = new XmlTextWriter(buffer);

                writer.WriteStartDocument();
                writer.WriteStartElement(SAS_REQUEST);
                writer.WriteElementString(ACTION, ACTION_PING);
                writer.WriteEndElement();
                writer.Close();

                Request = buffer.ToString();

                return DoRequest();
                
            } catch {
                return false;
            }
        }

        /// <summary>
        /// Check whether a user exists or not.
        /// </summary>
        /// <param name="username"></param>
        /// <returns></returns>
        public bool UserExists(string username) {
            try {
                // Create XML
                StringWriter buffer = new StringWriter();
                XmlTextWriter writer = new XmlTextWriter(buffer);

                writer.WriteStartDocument();
                writer.WriteStartElement(SAS_REQUEST);
                writer.WriteElementString(ACTION, ACTION_EXISTS);
                writer.WriteElementString(USERNAME, RemoveDomain(username));
                writer.WriteElementString(SECRET, Secret);
                writer.WriteEndElement();
                writer.Close();

                Request = buffer.ToString();

                bool result = DoRequest();

                return result;
            } catch (Exception e) {
                Debug.WriteLine(LOG_SOURCE, "Error checking user: {0}", e.Message);
                Error = e.Message;
                return false;
            }
        }

        /// <summary>Starts a new single channel session with the PINsafe 
        ///		server.</summary>
        /// <param name="username">The username for the user.</param>
        /// <returns>Session ID for the started session, or an empty string if 
        ///		an error occured.</returns>
        public string StartSession(string username) {
            string sessionID;

            try {
                // Create XML
                StringWriter buffer = new StringWriter();
                XmlTextWriter writer = new XmlTextWriter(buffer);

                writer.WriteStartDocument();
                writer.WriteStartElement(SAS_REQUEST);
                writer.WriteElementString(ACTION, ACTION_SESSION_START);
                writer.WriteElementString(USERNAME, RemoveDomain(username));
                writer.WriteElementString(SECRET, Secret);
                writer.WriteEndElement();
                writer.Close();

                Request = buffer.ToString();

                if (DoRequest()) {
                    // Parse out session ID
                    XmlNodeList sessionNodes = responseDoc.GetElementsByTagName(SESSION_ID);
                    sessionID = sessionNodes.Count > 0 ? sessionNodes[0].InnerText : "";
                } else {
                    sessionID = "";
                }
            } catch (Exception e) {
                Error = e.Message;
                sessionID = "";
            }

            return sessionID;
        }

        /// <summary>Perform an authentication with the PINsafe server.</summary>
        /// <param name="username">The username of the user.</param>
        /// <param name="password">The user's password.</param>
        /// <param name="otc">The user's one time code.</param>
        /// <param name="expiryTime">The time at which the login cookie will expire.</param>
        /// <returns>A secure cookie if login was successful. The cookie should be different depending on whether Change PIN is required or not. NULL if authentication fails.</returns>
        public string Login(string username, string password, string otc, DateTime expiryTime) {
            bool result;

            try {
                // If Non-PINsafe users are allowed, check if the user exists.
                bool doLogin = true;
                if (AllowNonPINsafeUsers) {
                    doLogin = UserExists(username);
                }

                // Create XML
                if (doLogin) {
                    StringWriter buffer = new StringWriter();
                    XmlTextWriter writer = new XmlTextWriter(buffer);

                    writer.WriteStartDocument();
                    writer.WriteStartElement(SAS_REQUEST);
                    writer.WriteElementString(ACTION, ACTION_LOGIN);
                    writer.WriteElementString(USERNAME, RemoveDomain(username));
                    writer.WriteElementString(PASSWORD, password);
                    writer.WriteElementString(OTC, otc);
                    writer.WriteElementString(SECRET, Secret);
                    writer.WriteEndElement();
                    writer.Close();

                    Request = buffer.ToString();

                    // Attempt authentication
                    result = DoRequest();
                    if (String.IsNullOrEmpty(Error)) {
                        Error = result
                                    ? "PINsafe login successful for " + username
                                    : "PINsafe login failed for " + username;
                    }
                } else {
                    // By-pass PINsafe login.
                    Error = "User " + username + " not known to PINsafe";
                    result = true;
                }
            } catch (Exception e) {
                Error = e.Message;
                result = false;
            }

            return (result) ? GetCookie(username, CookieType.PINsafe, expiryTime) : null;
        }

        private string GetCookie(string username, CookieType cookieType, DateTime expiryTime) {
            TimeSpan expirySpan = expiryTime.Subtract(new DateTime(1970, 1, 1));
            string expirySecs = ((long)expirySpan.TotalSeconds).ToString("D10");
            string timeAndUser = expirySecs + username + ":" + cookieType;
            return GetHashValue(timeAndUser) + timeAndUser;
        }

        public string GetSingleFactorCookie(string username, DateTime expiryTime) {
            return GetCookie(username, CookieType.Single, expiryTime);
        }

        public string GetPINsafeCookie(string username, DateTime expiryTime) {
            return GetCookie(username, CookieType.PINsafe, expiryTime);
        }

        public string GetChangePINCookie(string username, DateTime expiryTime) {
            return GetCookie(username, CookieType.ChangePIN, expiryTime);
        }

        public string UpdateCookie(string oldCookie, CookieType type) {
            if (IsCookieValid(oldCookie, type)) {
                string sExpiryTime = oldCookie.Substring(32, 10);
                DateTime oldExpiryTime = new DateTime(1970, 1, 1).AddSeconds(long.Parse(sExpiryTime));
                TimeSpan timeToExpire = oldExpiryTime - DateTime.Now;
                if (timeToExpire.TotalSeconds < settings.IdleTimeSecs - 10) {
                    // Cookie is more than 10 seconds old - renew it.
                    string username = oldCookie.Substring(42);
                    DateTime expiryTime = DateTime.Now.AddSeconds(settings.IdleTimeSecs);
                    return GetCookie(username, type, expiryTime);
                }
            }
            return null;
        }

        private string GetHashValue(string timeAndUser) {
            string data = CookieSecret + timeAndUser;
            byte[] md5 = MD5.Create().ComputeHash(Encoding.UTF8.GetBytes(data));
            StringBuilder hash = new StringBuilder();
            foreach (byte b in md5) {
                hash.Append(b.ToString("X2"));
            }
            return hash.ToString();
        }

        public bool CPCookieValid(string cookieData) {
            return IsCookieValid(cookieData, CookieType.ChangePIN);
        }

        public bool TwoFactorCookieValid(string cookieData) {
            return IsCookieValid(cookieData, CookieType.PINsafe);
        }

        public bool SingleFactorCookieValid(string cookieData) {
            return IsCookieValid(cookieData, CookieType.Single);
        }

        public string GetCookieUsername(string cookieData) {
            return IsCookieValid(cookieData, CookieType.PINsafe) ? GetUsernameFromCookie(cookieData) : null;
        }

        public string GetCPCookieUsername(string cookieData) {
            return IsCookieValid(cookieData, CookieType.ChangePIN) ? GetUsernameFromCookie(cookieData) : null;
        }

        private string GetUsernameFromCookie(string cookieData) {
            if (cookieData == null || cookieData.Length <= 42) {
                return null;
            }
            string username = cookieData.Substring(42);
            int colonIdx = username.IndexOf(':');
            if (colonIdx <= 0) {
                return "";
            }
            return username.Substring(0, colonIdx);
        }

        private bool IsCookieValid(string cookieData, CookieType type) {
            // Split the cookie up.
            if (cookieData == null || cookieData.Length <= 42) {
                return false;
            }
            string expiryString = cookieData.Substring(32, 10);
            long expiryLong;
            if (!long.TryParse(expiryString, out expiryLong)) {
                return false;
            }
            DateTime expiryTime = new DateTime(1970, 1, 1).AddSeconds(expiryLong);
            if (expiryTime < DateTime.Now) {
                return false; // Cookie has expired.
            }
            string orgHash = cookieData.Substring(0, 32);
            string username = cookieData.Substring(42);
            if (String.CompareOrdinal(orgHash, GetHashValue(expiryString + username)) != 0) {
                return false; // Invalid cookie.
            }
            int colonIdx = username.IndexOf(':');
            if (colonIdx < 0) {
                return false; // No username type
            }
            string cookieType = username.Substring(colonIdx + 1);
            return (String.Compare(cookieType, type.ToString(), true, CultureInfo.InvariantCulture) == 0);
        }

        /// <summary>Perform a PIN change with the PINsafe server.</summary>
        /// <param name="username">The username of the user.</param>
        /// <param name="password">The user's current password.</param>
        /// <param name="newPassword">The user's new password.</param>
        /// <param name="otc">The one time code for the user's current 
        ///		PIN.</param>
        /// <param name="newOtc">The one time code for the user's new 
        ///		PIN.</param>
        /// <returns>True if the PIN change was successful, otherwise 
        ///		false</returns>
        public bool ChangePin(string username, string password, string newPassword,
                              string otc, string newOtc) {
            bool result;

            try {
                // Create XML
                StringWriter buffer = new StringWriter();
                XmlTextWriter writer = new XmlTextWriter(buffer);

                writer.WriteStartDocument();
                writer.WriteStartElement(SAS_REQUEST);
                writer.WriteElementString(ACTION, ACTION_CHANGE_PIN);
                writer.WriteElementString(USERNAME, RemoveDomain(username));
                writer.WriteElementString(PASSWORD, password);
                writer.WriteElementString(NEW_PASSWORD, newPassword);
                writer.WriteElementString(OTC, otc);
                writer.WriteElementString(NEW_OTC, newOtc);
                writer.WriteElementString(SECRET, Secret);
                writer.WriteEndElement();
                writer.Close();

                Request = buffer.ToString();

                // Attempt pin change
                result = DoRequest();
            } catch (Exception e) {
                Error = e.Message;
                result = false;
            }

            return result;
        }

		public bool RequestResetCode(string username) {
			bool result;

			try {
				// Create XML
				StringWriter buffer = new StringWriter();
				XmlTextWriter writer = new XmlTextWriter(buffer);

				writer.WriteStartDocument();
				writer.WriteStartElement(SAS_REQUEST);
				writer.WriteElementString(ACTION, ACTION_RESET_CODE);
				writer.WriteElementString(USERNAME, RemoveDomain(username));
				writer.WriteElementString(SECRET, Secret);
				writer.WriteEndElement();
				writer.Close();

				Request = buffer.ToString();

				// Attempt pin change
				result = DoRequest();
			} catch (Exception e) {
				Error = e.Message;
				result = false;
			}

			return result;
		}

		public bool ResetPIN(string username, string resetCode) {
			bool result;

			try {
				// Create XML
				StringWriter buffer = new StringWriter();
				XmlTextWriter writer = new XmlTextWriter(buffer);

				writer.WriteStartDocument();
				writer.WriteStartElement(SAS_REQUEST);
				writer.WriteElementString(ACTION, ACTION_RESET);
				writer.WriteElementString(USERNAME, RemoveDomain(username));
				writer.WriteElementString(RESET_CODE, resetCode);
				writer.WriteElementString(SECRET, Secret);
				writer.WriteEndElement();
				writer.Close();

				Request = buffer.ToString();

				// Attempt pin change
				result = DoRequest();
			} catch (Exception e) {
				Error = e.Message;
				result = false;
			}

			return result;
		}

		public bool SendCustomMessage(string username, string message) {
			bool result;

			try {
				// Create XML
				StringWriter buffer = new StringWriter();
				XmlTextWriter writer = new XmlTextWriter(buffer);

				writer.WriteStartDocument();
				writer.WriteStartElement(HELPDESK_REQUEST);
				writer.WriteAttributeString(SECRET_ATTR, Secret);
				writer.WriteAttributeString(VERSION_ATTR, ADMIN_VERSION);
				writer.WriteStartElement(ADMIN_MESSAGE);
				writer.WriteAttributeString(REPOS_ATTR, "*");
				writer.WriteStartElement(ADMIN_USER);
				writer.WriteAttributeString(NAME_ATTR, username);
				writer.WriteStartElement(ADMIN_ALERT);
				writer.WriteAttributeString(TEXT_ATTR, message);
				writer.WriteEndElement();
				writer.WriteEndElement();
				writer.WriteEndElement();
				writer.WriteEndElement();
				writer.Close();

				Request = buffer.ToString();

				SendAdminRequest();
				// TODO: check that it's been sent.
				result = true;
			} catch (Exception e) {
				Error = e.Message;
				result = false;
			}

			return result;
		}

        /// <summary>Send an XML request to the PINsafe server.</summary>
        /// <param name="xml">The XML request.</param>
        /// <returns>The XML response from the server, or an empty string if 
        ///		an error occurred</returns>
        public string SendRequestXml(String xml) {
            Request = xml;
            SendAgentRequest();
            return Response;
        }

        /// <summary>Send the request to the PINsafe server and parse the result.</summary>
        /// <returns>True if the request was successful, false otherwise.</returns>
        private bool DoRequest() {
            bool result;

            SendAgentRequest();

            // Parse response
            responseDoc = new XmlDocument();
            responseDoc.LoadXml(Response);
            XmlNodeList resultNodes = responseDoc.GetElementsByTagName(RESULT);

            if ((resultNodes.Count == 1) && resultNodes[0].InnerText.Equals(PASS)) {
                result = true;

                // Parse out any warnings
                XmlNodeList warningNodes = responseDoc.GetElementsByTagName(WARNING);
                Warnings = new string[warningNodes.Count];
                for (int i = 0; i < warningNodes.Count; i++) {
                    Warnings[i] = warningNodes[i].InnerText;
                }
            } else {
                result = false;

                // Parse out the error
                XmlNodeList errorNodes = responseDoc.GetElementsByTagName(ERROR);
                if (errorNodes.Count > 0) {
                    Error = errorNodes[0].InnerText;
                }
            }
            return result;
        }

	    /// <summary>Send the request to the PINsafe server.</summary>
	    private void SendAgentRequest() {
			SendRequest(AgentXmlAddress);
	    }

		/// <summary>Send the request to the PINsafe server.</summary>
		private void SendAdminRequest() {
			SendRequest(AdminXmlAddress);
		}

		private void SendRequest(string address) {
			Response = "";

            byte[] requestData = Encoding.UTF8.GetBytes(HttpUtility.UrlEncode(Request) ?? "");

            WebClient client = new WebClient();
            if (AcceptSelfSigned) {
                ServicePointManager.ServerCertificateValidationCallback = DummyCertificateValidator;
            }
            byte[] responseData = client.UploadData(address, requestData);

            Response = Encoding.UTF8.GetString(responseData);
        }

        private bool DummyCertificateValidator(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors) {
            return true;
        }

        private static Image BlankImage {
            get {
                Bitmap img = new Bitmap(1, 1);
                img.SetPixel(0, 0, Color.White);
                return img;
            }
        }

        private Image GetImage(String sessionID, String username, string baseAddress, out int status) {
            if (string.IsNullOrEmpty(sessionID)) {
                if (!string.IsNullOrEmpty(username)) {
                    sessionID = StartSession(username);
                } else {
                    status = (int)HttpStatusCode.OK;
                    return BlankImage;
                }
            }

            if (string.IsNullOrEmpty(sessionID)) {
                sessionID = "invalid";
            }

            StringBuilder address = new StringBuilder(baseAddress);
            address.Append("?sessionid=");
            address.Append(sessionID);
            address.Append("&random=");
            address.Append(new Random().Next());
            string contentType;
            HttpStatusCode statusCode;
            Image image = GetPinsafeResponse(address.ToString(), out contentType, out statusCode);
            status = (int)statusCode;
            if (statusCode != HttpStatusCode.OK) {
                return BlankImage;
            }
            return image;
        }

        public Image GetImage(String sessionID, String username, out int status) {
            return GetImage(sessionID, username, ImageAddress, out status);
        }

        public Image GetMessage(String sessionID, String username, out int status) {
            return GetImage(sessionID, username, MessageAddress, out status);
        }

        /// <summary>
        /// Send a request to the PINSafe server, and get the image data.
        /// </summary>
        /// <param name="requestUrl">The URL of the PINSafe request.</param>
        /// <param name="contentType">The content type of the returned image.</param>
        /// <param name="status">The response code from the PINsafe server.</param>
        /// <returns>The PINSafe image data.</returns>
        private Image GetPinsafeResponse(string requestUrl, out string contentType, out HttpStatusCode status) {
            HttpWebRequest webRequest = (HttpWebRequest)WebRequest.Create(requestUrl);
            if (AcceptSelfSigned) {
                ServicePointManager.ServerCertificateValidationCallback = DummyCertificateValidator;
            }
            HttpWebResponse webResponse = (HttpWebResponse)webRequest.GetResponse();
            status = webResponse.StatusCode;
            if (status == HttpStatusCode.OK) {
                MemoryStream strm = new MemoryStream();
                Stream respStream = webResponse.GetResponseStream();
                if (respStream != null) {
                    const int bufSize = 0x400;
                    byte[] buffer = new byte[bufSize];
                    int bufLen = respStream.Read(buffer, 0, bufSize);
                    while (bufLen > 0) {
                        strm.Write(buffer, 0, bufLen);
                        bufLen = respStream.Read(buffer, 0, bufSize);
                    }
                }
                contentType = webResponse.ContentType;
                return Image.FromStream(strm);
            }
            contentType = "";
            return null;
        }

        private static string RemoveDomain(string username) {
            if (username == null) {
                return null;
            }
            int slashPos = username.IndexOf("\\", StringComparison.Ordinal);
            if (slashPos > 0 && slashPos < username.Length) {
                return username.Substring(slashPos + 1);
            }
            int atPos = username.IndexOf('@');
            return atPos > 0 ? username.Substring(0, atPos) : username;
        }
    }
}
