using System;
using System.Collections.Generic;
using System.Configuration;
using System.Security.Cryptography;
using System.Globalization;
using System.IO;
using System.Text;
using System.Xml;

namespace com.swivelsecure.pinsafe {

    /// <summary>
    /// Summary description for PINsafeSettings
    /// </summary>
    public class PINsafeSettings {

        private const string PINSAFE_START_COMMENT = "PINsafe settings";
        private const string PINSAFE_END_COMMENT = "End of PINsafe settings";
        private const string DEFAULT_LOGIN_URL = "/pinsafe/Login.aspx";
        private const string DEFAULT_LOGOFF_URL = "/pinsafe/Logoff.aspx";
        private const string DEFAULT_CHANGE_PIN_URL = "/pinsafe/ChangePin.aspx";
        private const string DEFAULT_CHANGE_PASSWORD_URL = "/pinsafe/ChangePW.aspx";
        private const string DEFAULT_AUTH_URL = "/_layouts/Authenticate.aspx";
        private const string DEFAULT_EXCLUSIONS = "/pinsafe/;/favicon.ico";
        private const string SWIVEL_DIR = "Swivel Secure";
        private const string APPLICATION_DIR = "PINsafe SharePoint Filter";
        private const string STORAGE_FILE = "PINsafeSettings.data";
        private const string STORAGE_DIR = "\\" + SWIVEL_DIR + "\\" + APPLICATION_DIR + "\\";
        public const string SHARE_NAME = "PINsafeSP";
        public const string SHARE_SEARCH_PATH = "Name='" + SHARE_NAME + "'";
        private const string CONFIG_SHARE_PATH = @"\\{0}\" + SHARE_NAME + "\\{1}";
        private const string PINSAFE_SERVER = "PINsafeServer";
        private const string PINSAFE_PORT = "PINsafePort";
        private const string PINSAFE_CONTEXT = "PINsafeContext";
        private const string PINSAFE_SECRET = "PINsafeSecret";
        private const string PINSAFE_SECURE = "PINsafeSecure";
        private const string PINSAFE_PASSWORD = "PINsafePassword";
        private const string PINSAFE_IMAGE = "PINsafeImage";
        private const string PINSAFE_MESSAGE = "PINsafeMessage";
        private const string PINSAFE_IGNORE_CERT_ERRORS = "PINsafeAcceptSelfSigned";
        private const string PINSAFE_COOKIE_SECRET = "PINsafeCookieSecret";
        private const string PINSAFE_EXCLUDED_PATHS = "PINsafeExcludedPaths";
        private const string PINSAFE_LOGON_URL = "PINsafeLogonPath";
        private const string PINSAFE_LOGOFF_URL = "PINsafeLogoffPath";
        private const string PINSAFE_CHANGE_PIN_URL = "PINsafeChangePinPath";
        private const string PINSAFE_CHANGE_PASSWORD_URL = "PINsafeChangePwdPath";
        private const string PINSAFE_AUTH_URL = "PINsafeAuthUrl";
        private const string ALLOW_NON_PINSAFE_USERS = "AllowNonPINsafeUsers";
        private const string PINSAFE_FILTER_ENABLED = "PINsafeFilterEnabled";
        private const string PINSAFE_REDIRECT_URL = "RedirectUrl";
        private const string PINSAFE_ROLE_PROVIDER = "PINsafeRoleProvider";
        private const string PINSAFE_AUTHENTICATION_DOMAIN = "PINsafeAuthenticationDomain";
        private const string SHAREPOINT_WEB_APPLICATION = "SharepointWebApplication";
        private const string PINSAFE_SETTINGS = "PINsafeSettings";
        private const string PINSAFE_IDLE_TIME_SECS = "PINsafeIdleTimeSecs";
        private const string PINSAFE_MEMBERSHIP_PROVIDER = "PINsafeMembershipProvider";
        private const string USERNAME_TYPE = "UsernameType";
        private const string LDAP_USERNAME = "UserAccount";
        private const string LDAP_PASSWORD = "UserPassword";
        private const string ENCRYPTION_KEY_VALUE = "mEXZAmP0QNV2nnltIazJ5w8Rb1jT2QfR/T+02YrpwaA=";
        private const string ENCRYPTION_IV_VALUE = "mGelJf0sglIPHxOsn413Tg==";
        private const string REQUIRES_PINSAFE = "RequiresPINsafeAuthentication";
        private const string DEBUG_ENABLED = "PINsafeDebugEnabled";
        private const string DEBUG_SOURCE = "PINsafeDebugSource";
	    private const string RESET_PIN_ENABLED = "ResetPinEnabled";
	    private const string RESET_PWD_ENABLED = "ResetPasswordEnabled";
	    private const string MANAGE_AD_PWDS = "ManageADPasswords";

        public enum UsernameTypes {
            sAMAccountName,
            userPrincipalName
        };

        public static string CommonDataPath {
            get { 
                string path = Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData);
                path += STORAGE_DIR;
                if (!Directory.Exists(path)) {
                    Directory.CreateDirectory(path); // Hopefully, admins have the correct access rights.
                }
                return path;
            }
        }

        public static bool ConfigFileExists(string filename) {
            return String.IsNullOrEmpty(filename) ? Directory.Exists(CommonDataPath) : File.Exists(CommonDataPath + filename);
        }

        public static string FailReason { get; private set; }

        public static Stream OpenConfigFileForRead(string filename) {
            try {
                return new FileStream(CommonDataPath + filename, FileMode.Open, FileAccess.Read);
            } catch (Exception e) {
                FailReason = e.Message;
                return null;
            }
        }

        public static Stream OpenConfigFileForWrite(string filename) {
            try {
                return new FileStream(CommonDataPath + filename, FileMode.Create);
            } catch (Exception e) {
                FailReason = e.Message;
                return null;
            }
        }

        public static Stream OpenConfigFileForAppend(string filename) {
            try {
                return new FileStream(CommonDataPath + filename, FileMode.Append);
            } catch (Exception e) {
                FailReason = e.Message;
                return null;
            }
        }

        public static bool SharedConfigFileExists(string serverName, string filename) {
            return (String.IsNullOrEmpty(filename)) ?
                Directory.Exists(String.Format(CONFIG_SHARE_PATH, serverName, "")) :
                File.Exists(String.Format(CONFIG_SHARE_PATH, serverName, filename));
        }

        public static Stream OpenSharedConfig(string serverName, string filename) {
            try {
                return new FileStream(String.Format(CONFIG_SHARE_PATH, serverName, filename), FileMode.Open, FileAccess.Read);
            } catch (Exception) {
                return null;
            }
        }

        // Server settings
        private bool ssl;
        private int port;
        private bool showPassword;
        private bool showTURing;
        private bool showMessage;
        private bool acceptSelfSigned;
        private List<string> excludedPaths;
        private long idleTimeSecs;
        private bool allowNonPINsafeUsers;
        private bool filterEnabled;
        private byte[] encryptionKey;
        private byte[] encryptionIV;
        private bool? requiresPINsafe;
	    private bool debugEnabled;
	    private bool resetPinEnabled;
	    private bool resetPwdEnabled;
	    private bool manageADPasswords;

        public DebugLogger Logger { get; private set; }

        public static PINsafeSettings SettingsFromConfiguration {
            get {
                PINsafeSettings settings = EmptySettings;
                settings.LoadSettingsFromConfiguration();
                return settings;
            }
        }

        public static PINsafeSettings SettingsFromLocation(string location) {
            PINsafeSettings settings = EmptySettings;
            return settings.LoadSettings(location) ? settings : null;
        }

        public static PINsafeSettings SettingsFromStorage {
            get {
                PINsafeSettings settings = EmptySettings;
                settings.ReadSettingsFromStorage();
                return settings;
            }
        }

        public static PINsafeSettings SettingsFromShared(string serverName) {
            PINsafeSettings settings = EmptySettings;
            return (settings.ReadSettingsFromSharedConfig(serverName) ? settings : null);
        }

        public static PINsafeSettings EmptySettings {
            get { return new PINsafeSettings(); }
        }

        private PINsafeSettings() { }

        public bool Ssl {
            get { return ssl; }
            set { ssl = value; }
        }

        public string Server { get; set; }

        public int Port {
            get { return port; }
            set { port = value; }
        }

        public string Context { get; set; }

        public string Secret { get; set; }

        public bool ShowPassword {
            get { return showPassword; }
            set { showPassword = value; }
        }

        public bool ShowTURing {
            get { return showTURing; }
            set { showTURing = value; }
        }

        public bool ShowMessage {
            get { return showMessage; }
            set { showMessage = value; }
        }

        public bool AcceptSelfSigned {
            get { return acceptSelfSigned; }
            set { acceptSelfSigned = value; }
        }

        public string CookieSecret { get; set; }

        public string[] ExcludedPaths {
            get { return excludedPaths.ToArray(); }
            set {
                excludedPaths = new List<string>(value);
            }
        }

        public string LogonUrl { get; set; }

        public string LogoffUrl { get; set; }

        public string ChangePinUrl { get; set; }

        public string ChangePasswordUrl { get; set; }

        public string AuthUrl { get; set; }

        public string RedirectUrl { get; set; }

        public long IdleTimeSecs {
            get { return idleTimeSecs; }
            set { idleTimeSecs = value; }
        }

        public bool AllowNonPINsafeUsers {
            get { return allowNonPINsafeUsers; }
            set { allowNonPINsafeUsers = value; }
        }

        public bool FilterEnabled {
            get { return filterEnabled; }
            set { filterEnabled = value; }
        }

        public string MembershipProvider { get; set; }

        public string AuthenticationDomain { get; set; }

        public string RoleProvider { get; set; }

        public string WebApplication { get; set; }

        public UsernameTypes UsernameType { get; set; }

        public string UserAccount { get; set; }

        public string UserPassword { get; set; }

        public bool? RequiresPINsafe { get { return requiresPINsafe; } }

	    public bool DebugEnabled {
		    get { return debugEnabled; }
		    set { debugEnabled = value; }
	    }

	    public bool ResetPinEnabled {
			get { return resetPinEnabled; }
			set { resetPinEnabled = value; }
	    }

	    public bool ResetPasswordEnabled {
			get { return resetPwdEnabled; }
			set { resetPwdEnabled = value; }
	    }

	    public bool ManageADPasswords {
			get { return manageADPasswords; }
			set { manageADPasswords = value; }
	    }

	    public void LoadSettingsFromConfiguration() {
		    Server = ConfigurationManager.AppSettings[PINSAFE_SERVER];
		    String sPort = ConfigurationManager.AppSettings[PINSAFE_PORT];
		    Context = ConfigurationManager.AppSettings[PINSAFE_CONTEXT];
		    Secret = ConfigurationManager.AppSettings[PINSAFE_SECRET];
		    String sSsl = ConfigurationManager.AppSettings[PINSAFE_SECURE];
		    String showPw = ConfigurationManager.AppSettings[PINSAFE_PASSWORD];
		    String showImg = ConfigurationManager.AppSettings[PINSAFE_IMAGE];
		    String showMsg = ConfigurationManager.AppSettings[PINSAFE_MESSAGE];
		    String allowSelfSigned = ConfigurationManager.AppSettings[PINSAFE_IGNORE_CERT_ERRORS];
		    CookieSecret = ConfigurationManager.AppSettings[PINSAFE_COOKIE_SECRET];
		    string excPathString = ConfigurationManager.AppSettings[PINSAFE_EXCLUDED_PATHS];
		    ExcludedPaths = excPathString == null ? new string[0] : excPathString.Split(';');
		    LogonUrl = ConfigurationManager.AppSettings[PINSAFE_LOGON_URL];
		    LogoffUrl = ConfigurationManager.AppSettings[PINSAFE_LOGOFF_URL];
		    ChangePinUrl = ConfigurationManager.AppSettings[PINSAFE_CHANGE_PIN_URL];
		    ChangePasswordUrl = ConfigurationManager.AppSettings[PINSAFE_CHANGE_PASSWORD_URL];
		    AuthUrl = ConfigurationManager.AppSettings[PINSAFE_AUTH_URL];
		    String sIdleTimeSecs = ConfigurationManager.AppSettings[PINSAFE_IDLE_TIME_SECS];
		    String sAllowNonPINsafeUsers = ConfigurationManager.AppSettings[ALLOW_NON_PINSAFE_USERS];
		    String sFilterEnabled = ConfigurationManager.AppSettings[PINSAFE_FILTER_ENABLED];
		    RedirectUrl = ConfigurationManager.AppSettings[PINSAFE_REDIRECT_URL];
		    MembershipProvider = ConfigurationManager.AppSettings[PINSAFE_MEMBERSHIP_PROVIDER];
		    RoleProvider = ConfigurationManager.AppSettings[PINSAFE_ROLE_PROVIDER];
		    AuthenticationDomain = ConfigurationManager.AppSettings[PINSAFE_AUTHENTICATION_DOMAIN];
		    WebApplication = ConfigurationManager.AppSettings[SHAREPOINT_WEB_APPLICATION];
		    UsernameType = GetUsernameType(ConfigurationManager.AppSettings[USERNAME_TYPE]);
		    UserAccount = ConfigurationManager.AppSettings[LDAP_USERNAME];
		    encryptionKey = Convert.FromBase64CharArray(ENCRYPTION_KEY_VALUE.ToCharArray(), 0, ENCRYPTION_KEY_VALUE.Length);
		    encryptionIV = Convert.FromBase64CharArray(ENCRYPTION_IV_VALUE.ToCharArray(), 0, ENCRYPTION_IV_VALUE.Length);
		    UserPassword = Decrypt(ConfigurationManager.AppSettings[LDAP_PASSWORD]);
		    String sRequiresPINsafe = ConfigurationManager.AppSettings[REQUIRES_PINSAFE];

		    if (sSsl == null || sSsl.Equals("")) {
			    ssl = false;
		    } else {
			    if (!Boolean.TryParse(sSsl, out ssl)) {
				    ssl = false;
			    }
		    }
		    if (sPort == null || sPort.Equals("")) {
			    port = (Ssl ? 8443 : 8080);
		    } else {
			    if (!Int32.TryParse(sPort, out port)) {
				    port = (Ssl ? 8443 : 8080);
			    }
		    }
		    if (!Boolean.TryParse(showPw, out showPassword)) {
			    showPassword = true;
		    }
		    if (!Boolean.TryParse(showImg, out showTURing)) {
			    showTURing = true;
		    }
		    if (!Boolean.TryParse(showMsg, out showMessage)) {
			    showMessage = false;
		    }
		    if (!Boolean.TryParse(allowSelfSigned, out acceptSelfSigned)) {
			    acceptSelfSigned = false;
		    }
		    if (!long.TryParse(sIdleTimeSecs, out idleTimeSecs)) {
			    idleTimeSecs = 300; // Default 5 minutes.
		    }
		    if (!Boolean.TryParse(sAllowNonPINsafeUsers, out allowNonPINsafeUsers)) {
			    allowNonPINsafeUsers = false;
		    }
		    if (!Boolean.TryParse(sFilterEnabled, out filterEnabled)) {
			    filterEnabled = true;
		    }
		    bool bRequiresPINsafe;
		    if (String.IsNullOrEmpty(sRequiresPINsafe)) {
			    requiresPINsafe = null;
		    } else if (Boolean.TryParse(sRequiresPINsafe, out bRequiresPINsafe)) {
			    requiresPINsafe = bRequiresPINsafe;
		    } else {
			    requiresPINsafe = null;
		    }
		    string sDebugEnabled = ConfigurationManager.AppSettings[DEBUG_ENABLED];
		    debugEnabled = true;
		    if (!String.IsNullOrEmpty(sDebugEnabled) && !Boolean.TryParse(sDebugEnabled, out debugEnabled)) {
			    debugEnabled = true;
		    }
		    //string debugSource = ConfigurationManager.AppSettings[DEBUG_SOURCE] ?? "Swivel";
		    Logger = new DebugLogger(debugEnabled);

			string pwdManage = ConfigurationManager.AppSettings[MANAGE_AD_PWDS];
			if (String.IsNullOrEmpty(pwdManage)) {
				manageADPasswords = true;
			} else if (!Boolean.TryParse(pwdManage, out manageADPasswords)) {
				manageADPasswords = true;
			}
			string sResetEnabled = ConfigurationManager.AppSettings[RESET_PIN_ENABLED];
		    if (String.IsNullOrEmpty(sResetEnabled)) {
			    resetPinEnabled = true;
		    } else {
			    if (!Boolean.TryParse(sResetEnabled, out resetPinEnabled)) {
				    resetPinEnabled = false;
			    }
		    }
		    if (resetPinEnabled) {
			    if (ManageADPasswords) {
				    sResetEnabled = ConfigurationManager.AppSettings[RESET_PWD_ENABLED];
				    if (!String.IsNullOrEmpty(sResetEnabled)) {
					    if (!Boolean.TryParse(sResetEnabled, out resetPwdEnabled)) {
						    resetPwdEnabled = false;
					    }
				    } else {
					    resetPwdEnabled = true;
				    }
			    } else {
				    resetPwdEnabled = false;
			    }
		    } else {
			    resetPwdEnabled = false;
		    }
	    }

	    private static UsernameTypes GetUsernameType(string usernameType) {
            UsernameTypes type;
            if (String.IsNullOrEmpty(usernameType)) {
                type = UsernameTypes.sAMAccountName;
            } else {
                try {
                    type = (UsernameTypes) Enum.Parse(typeof (UsernameTypes), usernameType);
                } catch (ArgumentException) {
                    type = UsernameTypes.sAMAccountName;
                }
            }
            return type;
        }

        public void WriteSettingsToStorage() {
            XmlDocument configDoc = new XmlDocument();
            XmlElement settingsElem = configDoc.CreateElement(PINSAFE_SETTINGS);
            configDoc.AppendChild(settingsElem);
            IDictionary<string, string> settings = new Dictionary<string, string>();
            SettingsToXml(configDoc, settingsElem, settings);
            using (Stream stream = OpenConfigFileForWrite(STORAGE_FILE)) {
                configDoc.Save(stream);
            }
        }

        public void ReadSettingsFromStorage() {
            if (!ConfigFileExists(STORAGE_FILE)) {
                // File doesn't exist - use defaults.
                SetDefaults();
                return;
            }
            XmlDocument configDoc = new XmlDocument();
            using (Stream stream = OpenConfigFileForRead(STORAGE_FILE)) {
                if (stream == null) {
                    SetDefaults();
                    return;
                }
                try {
                    configDoc.Load(stream);
                } catch {
                    // Not a valid XML document, or file does not exist.
                    SetDefaults();
                    return;
                }
            }
            XmlElement settingsElem = configDoc.DocumentElement;
            if (settingsElem == null) {
                SetDefaults();
                return;
            }
            SettingsFromXml(settingsElem);
        }

        public bool ReadSettingsFromSharedConfig(string serverName) {
            XmlDocument configDoc = new XmlDocument();
            using (Stream stream = OpenSharedConfig(serverName, STORAGE_FILE)) {
                if (stream == null) {
                    ReadSettingsFromStorage();
                    return false;
                }
                try {
                    configDoc.Load(stream);
                } catch {
                    // Not a valid XML document, or file does not exist.
                    SetDefaults();
                    return false;
                }
            }
            XmlElement settingsElem = configDoc.DocumentElement;
            if (settingsElem == null) {
                SetDefaults();
                return false;
            }
            SettingsFromXml(settingsElem);
            return true;
        }

        private static Dictionary<string, string> ReadSettings(XmlElement parent) {
            Dictionary<string, string> settings = new Dictionary<string, string>();
            foreach (XmlElement elem in parent.GetElementsByTagName("add")) {
                settings[elem.GetAttribute("key")] = elem.GetAttribute("value");
            }
            return settings;
        }

        private static string GetSetting(IDictionary<string, string> settings, string name, string defValue = "") {
            return (settings.ContainsKey(name)) ? settings[name] : defValue;
        }

        private static bool GetBoolSetting(IDictionary<string, string> settings, string name, bool defValue) {
            if (settings.ContainsKey(name)) {
                string sValue = settings[name];
                bool value;
                if (bool.TryParse(sValue, out value)) {
                    return value;
                }
            }
            return defValue;
        }

        private static bool? GetOptionalBoolSetting(IDictionary<string, string> settings, string name) {
            if (settings.ContainsKey(name)) {
                string sValue = settings[name];
                bool value;
                if (bool.TryParse(sValue, out value)) {
                    return value;
                }
            }
            return null;
        }

        private static int GetIntSetting(IDictionary<string, string> settings, string name, int defValue) {
            if (settings.ContainsKey(name)) {
                string sValue = settings[name];
                int value;
                if (int.TryParse(sValue, out value)) {
                    return value;
                }
            }
            return defValue;
        }

        public bool LoadSettings(string location) {
            if (File.Exists(location)) {
                using (StreamReader reader = new StreamReader(location)) {
                    XmlDocument configDoc = new XmlDocument();
                    try {
                        configDoc.Load(reader);
                    } catch {
                        return false;
                    }
                    if (configDoc.DocumentElement != null) {
                        XmlElement settingsElem = configDoc.DocumentElement["appSettings"];
                        SettingsFromXml(settingsElem);
                    }
                }
            } else {
                SetDefaults();
            }
            return true;
        }

        private void SettingsFromXml(XmlElement settingsElem) {
            Dictionary<string, string> settings = ReadSettings(settingsElem);
            Server = GetSetting(settings, PINSAFE_SERVER);
            ssl = GetBoolSetting(settings, PINSAFE_SECURE, false);
            port = GetIntSetting(settings, PINSAFE_PORT, (ssl ? 8443 : 8080));
            Context = GetSetting(settings, PINSAFE_CONTEXT, "pinsafe");
            Secret = GetSetting(settings, PINSAFE_SECRET);
            showPassword = GetBoolSetting(settings, PINSAFE_PASSWORD, true);
            showTURing = GetBoolSetting(settings, PINSAFE_IMAGE, true);
            showMessage = GetBoolSetting(settings, PINSAFE_MESSAGE, false);
            acceptSelfSigned = GetBoolSetting(settings, PINSAFE_IGNORE_CERT_ERRORS, false);
            CookieSecret = GetSetting(settings, PINSAFE_COOKIE_SECRET);
            string sExcPaths = GetSetting(settings, PINSAFE_EXCLUDED_PATHS, DEFAULT_EXCLUSIONS);
            ExcludedPaths = (sExcPaths == null) ? new string[0] : sExcPaths.Split(';');
            LogonUrl = GetSetting(settings, PINSAFE_LOGON_URL, DEFAULT_LOGIN_URL);
            LogoffUrl = GetSetting(settings, PINSAFE_LOGOFF_URL, DEFAULT_LOGOFF_URL);
            ChangePinUrl = GetSetting(settings, PINSAFE_CHANGE_PIN_URL, DEFAULT_CHANGE_PIN_URL);
            ChangePasswordUrl = GetSetting(settings, PINSAFE_CHANGE_PASSWORD_URL, DEFAULT_CHANGE_PASSWORD_URL);
            AuthUrl = GetSetting(settings, PINSAFE_AUTH_URL, DEFAULT_AUTH_URL);
            idleTimeSecs = GetIntSetting(settings, PINSAFE_IDLE_TIME_SECS, 300);
            allowNonPINsafeUsers = GetBoolSetting(settings, ALLOW_NON_PINSAFE_USERS, false);
            filterEnabled = GetBoolSetting(settings, PINSAFE_FILTER_ENABLED, true);
            RedirectUrl = GetSetting(settings, PINSAFE_REDIRECT_URL);
            MembershipProvider = GetSetting(settings, PINSAFE_MEMBERSHIP_PROVIDER);
            RoleProvider = GetSetting(settings, PINSAFE_ROLE_PROVIDER);
            AuthenticationDomain = GetSetting(settings, PINSAFE_AUTHENTICATION_DOMAIN);
            WebApplication = GetSetting(settings, SHAREPOINT_WEB_APPLICATION);
            UsernameType = GetUsernameType(GetSetting(settings, USERNAME_TYPE));
            UserAccount = GetSetting(settings, LDAP_USERNAME);
            encryptionKey = Convert.FromBase64CharArray(ENCRYPTION_KEY_VALUE.ToCharArray(), 0, ENCRYPTION_KEY_VALUE.Length);
            encryptionIV = Convert.FromBase64CharArray(ENCRYPTION_IV_VALUE.ToCharArray(), 0, ENCRYPTION_IV_VALUE.Length);
            UserPassword = Decrypt(ConfigurationManager.AppSettings[LDAP_PASSWORD]);
            UserPassword = Decrypt(GetSetting(settings, LDAP_PASSWORD));
            requiresPINsafe = GetOptionalBoolSetting(settings, REQUIRES_PINSAFE);
	        debugEnabled = GetBoolSetting(settings, DEBUG_ENABLED, true);
	        resetPinEnabled = GetBoolSetting(settings, RESET_PIN_ENABLED, true);
			manageADPasswords = GetBoolSetting(settings, MANAGE_AD_PWDS, true);
			resetPwdEnabled = resetPinEnabled && manageADPasswords && GetBoolSetting(settings, RESET_PWD_ENABLED, true);
            Logger = new DebugLogger(debugEnabled);
        }

        private static void WriteSettings(XmlDocument configDoc, XmlNode parent, IDictionary<string, string> settings) {
            XmlNodeList nodes = parent.ChildNodes;
            XmlNode endComment = null;
            XmlText prevNode = null;
            XmlText lineBreak = null;
            for (int idx = 0; idx < nodes.Count; idx++) {
	            XmlElement elem = nodes[idx] as XmlElement;
	            if (elem != null) {
                    if (elem.Name.Equals("add")) {
                        string name = elem.GetAttribute("key");
                        if (settings.ContainsKey(name)) {
                            elem.SetAttribute("value", settings[name]);
                            settings.Remove(name);
                        } else {
                            parent.RemoveChild(elem);
                        }
                        if (prevNode != null) {
                            lineBreak = configDoc.CreateTextNode(prevNode.InnerText);
                        }
                    }
                } else if (nodes[idx] is XmlComment) {
                    string comment = nodes[idx].InnerText;
                    if (comment.ToLower().Contains(PINSAFE_END_COMMENT.ToLower())) {
                        endComment = nodes[idx];
                        if (prevNode != null) {
                            lineBreak = configDoc.CreateTextNode(prevNode.InnerText);
                        }
                    }
                } else if (lineBreak == null && nodes[idx] is XmlText) {
                    prevNode = (XmlText)nodes[idx];
                }
            }
	        if (lineBreak == null) {
                lineBreak = prevNode == null ? configDoc.CreateTextNode("\n    ") : configDoc.CreateTextNode(prevNode.InnerText);
            }
            if (endComment == null) {
                XmlNode lastNode = nodes[nodes.Count - 1];
                // Create comments to bracket the PINsafe entries.
                XmlComment startComment = configDoc.CreateComment(PINSAFE_START_COMMENT);
                parent.InsertBefore(lineBreak.CloneNode(true), lastNode);
                parent.InsertBefore(startComment, lastNode);
                endComment = configDoc.CreateComment(PINSAFE_END_COMMENT);
                parent.InsertBefore(lineBreak.CloneNode(true), lastNode);
                parent.InsertBefore(endComment, lastNode);
            }
            foreach (KeyValuePair<string, string> entry in settings) {
                XmlElement elem = configDoc.CreateElement("add");
                elem.SetAttribute("key", entry.Key);
                elem.SetAttribute("value", entry.Value);
                parent.InsertBefore(elem, endComment);
                parent.InsertBefore(lineBreak.CloneNode(true), endComment);
            }
        }

        public bool SaveSettings(string location, bool? requiresPS) {
            if (File.Exists(location)) {
                XmlDocument configDoc = new XmlDocument {PreserveWhitespace = true};
                using (StreamReader reader = new StreamReader(location)) {
                    try {
                        configDoc.Load(reader);
                    } catch {
                        return false;
                    }
                }
                XmlElement rootElem = configDoc.DocumentElement;
                if (rootElem == null) {
                    return false;
                }
                XmlElement settingsElem = rootElem["appSettings"];
                if (settingsElem == null) {
                    settingsElem = configDoc.CreateElement("appSettings");
                    rootElem.AppendChild(settingsElem);
                }

                Dictionary<string, string> settings = ReadSettings(settingsElem);
                if (requiresPS != null) {
                    settings[REQUIRES_PINSAFE] = ((bool)requiresPS).ToString(CultureInfo.InvariantCulture);
                } else if (requiresPINsafe != null) {
                    settings[REQUIRES_PINSAFE] = ((bool)requiresPINsafe).ToString(CultureInfo.InvariantCulture);
                }
                // Write all the settings to the XML file.
                SettingsToXml(configDoc, settingsElem, settings);
                // Make a backup.
                string backupName = location + ".bak";
                if (File.Exists(backupName)) {
                    File.Delete(backupName);
                }
                File.Move(location, backupName);
                using (StreamWriter writer = new StreamWriter(location)) {
                    XmlWriterSettings xSettings = new XmlWriterSettings {Indent = true, IndentChars = "\t"};
                    XmlWriter w = XmlWriter.Create(writer, xSettings);
                    try {
                        configDoc.Save(w);
                    } catch {
                        return false;
                    }
                }
            }
            return true;
        }

        private void SettingsToXml(XmlDocument configDoc, XmlElement settingsElem, IDictionary<string, string> settings) {
            settings[PINSAFE_SERVER] = Server;
            settings[PINSAFE_PORT] = Port.ToString(CultureInfo.InvariantCulture);
            settings[PINSAFE_CONTEXT] = Context;
            settings[PINSAFE_SECRET] = Secret;
            settings[PINSAFE_SECURE] = Ssl.ToString(CultureInfo.InvariantCulture);
            settings[PINSAFE_PASSWORD] = showPassword.ToString(CultureInfo.InvariantCulture);
            settings[PINSAFE_IMAGE] = showTURing.ToString(CultureInfo.InvariantCulture);
            settings[PINSAFE_MESSAGE] = showMessage.ToString(CultureInfo.InvariantCulture);
            settings[PINSAFE_IGNORE_CERT_ERRORS] = acceptSelfSigned.ToString(CultureInfo.InvariantCulture);
            if (String.IsNullOrEmpty(CookieSecret) || "System.Char[]".Equals(CookieSecret)) {
                CookieSecret = GenerateRandomCookie();
            }
            settings[PINSAFE_COOKIE_SECRET] = CookieSecret;
            if (excludedPaths != null && excludedPaths.Count > 0) {
                StringBuilder exPathString = new StringBuilder();
                foreach (string exPath in excludedPaths) {
                    exPathString.Append(exPath);
                    exPathString.Append(";");
                }
                if (exPathString.Length > 0) {
                    // Remove trailing semi-colon
                    exPathString.Remove(exPathString.Length - 1, 1);
                }
                settings[PINSAFE_EXCLUDED_PATHS] = exPathString.ToString();
            } else {
                settings[PINSAFE_EXCLUDED_PATHS] = "";
            }
            settings[PINSAFE_LOGON_URL] = LogonUrl;
            settings[PINSAFE_LOGOFF_URL] = LogoffUrl;
            settings[PINSAFE_CHANGE_PIN_URL] = ChangePinUrl;
            settings[PINSAFE_CHANGE_PASSWORD_URL] = ChangePasswordUrl;
            settings[PINSAFE_AUTH_URL] = AuthUrl;
            settings[PINSAFE_IDLE_TIME_SECS] = IdleTimeSecs.ToString(CultureInfo.InvariantCulture);
            settings[ALLOW_NON_PINSAFE_USERS] = AllowNonPINsafeUsers.ToString(CultureInfo.InvariantCulture);
            settings[PINSAFE_REDIRECT_URL] = AllowNonPINsafeUsers ? RedirectUrl : "";
            settings[PINSAFE_FILTER_ENABLED] = FilterEnabled.ToString(CultureInfo.InvariantCulture);
            settings[PINSAFE_MEMBERSHIP_PROVIDER] = MembershipProvider;
            settings[PINSAFE_ROLE_PROVIDER] = RoleProvider;
            settings[PINSAFE_AUTHENTICATION_DOMAIN] = AuthenticationDomain;
            settings[SHAREPOINT_WEB_APPLICATION] = WebApplication;
            settings[USERNAME_TYPE] = UsernameType.ToString();
            settings[LDAP_USERNAME] = UserAccount;
            if (encryptionKey == null || encryptionIV == null) {
                SymmetricAlgorithm alg = new AesCryptoServiceProvider();
                alg.GenerateKey();
                encryptionKey = alg.Key;
                alg.GenerateIV();
                encryptionIV = alg.IV;
            }
            settings[LDAP_PASSWORD] = Encrypt(UserPassword);
	        settings[DEBUG_ENABLED] = debugEnabled.ToString(CultureInfo.InvariantCulture);
	        settings[DEBUG_SOURCE] = "Swivel";
	        settings[RESET_PIN_ENABLED] = ResetPinEnabled.ToString(CultureInfo.InvariantCulture);
			settings[RESET_PWD_ENABLED] = ResetPasswordEnabled.ToString(CultureInfo.InvariantCulture);
	        settings[MANAGE_AD_PWDS] = ManageADPasswords.ToString(CultureInfo.InvariantCulture);
			WriteSettings(configDoc, settingsElem, settings);
        }

        public void SetDefaults() {
            // Configure sensible default values.
            Server = "localhost";
            port = 8080;
            Context = "pinsafe";
            Secret = "";
            ssl = false;
            showPassword = false;
            showTURing = true;
            showMessage = false;
            acceptSelfSigned = false;
            CookieSecret = GenerateRandomCookie();
            LogonUrl = DEFAULT_LOGIN_URL;
            LogoffUrl = DEFAULT_LOGOFF_URL;
	        ChangePinUrl = DEFAULT_CHANGE_PIN_URL;
	        ChangePasswordUrl = DEFAULT_CHANGE_PASSWORD_URL;
            AuthUrl = DEFAULT_AUTH_URL;
            ExcludedPaths = DEFAULT_EXCLUSIONS.Split(';');
            idleTimeSecs = 300;
            allowNonPINsafeUsers = false;
            RedirectUrl = "";
            filterEnabled = true;
            MembershipProvider = "ADMembership";
            RoleProvider = "";
            AuthenticationDomain = "";
	        resetPinEnabled = true;
	        resetPwdEnabled = true;
	        manageADPasswords = true;
        }

        public static void RemoveSettings(string configFile) {
            if (!File.Exists(configFile)) return;
            XmlDocument configDoc = new XmlDocument { PreserveWhitespace = true };
            using (StreamReader reader = new StreamReader(configFile)) {
                using (XmlTextReader xReader = new XmlTextReader(reader)) {
                    // We don't want to modify any of the document except for the bits that are relevant.
                    configDoc.Load(xReader);
                }
            }
            if (configDoc.DocumentElement == null) {
                return;
            }

            XmlElement settingsElem = configDoc.DocumentElement["appSettings"];
            if (settingsElem != null) {
                bool started = false;
                XmlNode node = settingsElem.FirstChild;
                while (node != null) {
                    if (node is XmlComment && node.InnerText.ToLower().Contains(PINSAFE_START_COMMENT.ToLower())) {
                        started = true;
                        XmlNode oldNode = node;
                        node = oldNode.NextSibling;
                        settingsElem.RemoveChild(oldNode);
                    } else if (started && node is XmlElement && node.Name == "add") {
                        XmlNode oldNode = node;
                        node = oldNode.NextSibling;
                        settingsElem.RemoveChild(oldNode);
                    } else if (started && node is XmlComment && node.InnerText.ToLower().Contains(PINSAFE_END_COMMENT.ToLower())) {
                        settingsElem.RemoveChild(node);
                        node = null;
                    } else {
                        node = node.NextSibling;
                    }
                }
            }
            // Make a backup.
            string backupName = configFile + ".bak";
            if (File.Exists(backupName)) {
                File.Delete(backupName);
            }
            File.Move(configFile, backupName);
            using (StreamWriter writer = new StreamWriter(configFile)) {
                configDoc.Save(writer);
            }
        }

        private static string ToBase64(byte[] input) {
            if (input == null || input.Length == 0) {
                return "";
            }
            int arrLen = (int)((4.0d / 3.0d) * input.Length);
            if (arrLen % 4 != 0) {
                arrLen += 4 - arrLen % 4;
            }
            char[] resultChars = new char[arrLen];
            Convert.ToBase64CharArray(input, 0, input.Length, resultChars, 0);
            return new String(resultChars);
        }

        private string Decrypt(string value) {
            if (string.IsNullOrEmpty(value)) {
                return "";
            }
            if (encryptionKey == null || encryptionIV == null) {
                // No encryption settings: use plain text.
                return value;
            }
            try {
                ICryptoTransform decryptor = new AesCryptoServiceProvider().CreateDecryptor(encryptionKey, encryptionIV);
                byte[] encrypted = Convert.FromBase64CharArray(value.ToCharArray(), 0, value.Length);
                using (MemoryStream strm = new MemoryStream(encrypted)) {
                    using (CryptoStream cstrm = new CryptoStream(strm, decryptor, CryptoStreamMode.Read)) {
                        using (StreamReader reader = new StreamReader(cstrm, Encoding.UTF8)) {
                            return reader.ReadToEnd();
                        }
                    }
                }
            } catch {
                return value;
            }
        }

        private string Encrypt(string value) {
            if (String.IsNullOrEmpty(value)) {
                return "";
            }
            SymmetricAlgorithm alg = new AesCryptoServiceProvider();
            try {
                ICryptoTransform encryptor = alg.CreateEncryptor(encryptionKey, encryptionIV);
                using (MemoryStream strm = new MemoryStream()) {
                    using (CryptoStream cstrm = new CryptoStream(strm, encryptor, CryptoStreamMode.Write)) {
                        using (StreamWriter writer = new StreamWriter(cstrm, Encoding.UTF8)) {
                            writer.Write(value);
                        }
                        return ToBase64(strm.ToArray());
                    }
                }
            } catch {
                return value;
            }
        }

        private static String GenerateRandomCookie(int length = 0x30) {
            char[] cookie = new char[length];
            Random rnd = new Random();
            for (int idx = 0; idx < length; idx++) {
                cookie[idx] = (char)(rnd.Next(0x20, 0x60));
            }
            return new string(cookie);
        }
    }
}