﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml;

namespace com.swivelsecure.pinsafe {
	public class PasswordGenerator {

		public const string SETTINGS_FILE = "PasswordSettings.xml";
		private const string CHAR_SETS_ELEM = "CharacterSets";
		private const string CHAR_SET_ELEM = "CharacterSet";
		private const string NAME_ATTR = "name";
		private const string TEMPLATES_ELEM = "Templates";
		private const string TEMPLATE_ELEM = "Template";
		private const string LOG_SOURCE = "PASSWORD_GENERATOR";

		private readonly string[] DEFAULT_CHARACTER_SETS = new[] {
			"ABCDEFGHJKLMNPQRSTUVWXYZ",
			"abcdefghjkmnpqrstuvwxyz",
			"1234567890"
		};

		private const string DEFAULT_TEMPLATE = "ULLLLDDD";

		private readonly Dictionary<Char, String> characterSets;
		private readonly List<String> templates;
		private readonly DebugLogger logger;

		private static PasswordGenerator instance;

		public static PasswordGenerator Instance {
			get { return instance ?? (instance = new PasswordGenerator()); }
		}

		private PasswordGenerator() {
			characterSets = new Dictionary<char, string>();
			templates = new List<string>();
			logger = PINsafeSettings.SettingsFromConfiguration.Logger;
			LoadSettings();
		}

		private void LoadSettings() {
			characterSets.Clear();
			templates.Clear();
			if (PINsafeSettings.ConfigFileExists(SETTINGS_FILE)) {
				XmlDocument doc = new XmlDocument();
				using (Stream settings = PINsafeSettings.OpenConfigFileForRead(SETTINGS_FILE)) {
					try {
						doc.Load(settings);
						XmlElement root = doc.DocumentElement;
						if (root != null) {
							XmlElement charSetsElem = root[CHAR_SETS_ELEM];
							if (charSetsElem != null) {
								foreach (XmlElement charSetElem in charSetsElem.GetElementsByTagName(CHAR_SET_ELEM)) {
									string name = charSetElem.GetAttribute(NAME_ATTR);
									string value = charSetElem.InnerText;
									if (!String.IsNullOrEmpty(name) && !String.IsNullOrEmpty(value)) {
										characterSets[name[0]] = value;
										logger.WriteLine(LOG_SOURCE, "Character set {0} = \"{1}\"", name[0], value);
									}
								}
							}
							XmlElement templatesElem = root[TEMPLATES_ELEM];
							if (templatesElem != null) {
								foreach (XmlElement tempElem in templatesElem.GetElementsByTagName(TEMPLATE_ELEM).Cast<XmlElement>().Where(tempElem => !String.IsNullOrEmpty(tempElem.InnerText))) {
									templates.Add(tempElem.InnerText);
									logger.WriteLine(LOG_SOURCE, "Template added: \"{0}\"", tempElem.InnerText);
								}
							}
						}
					} catch {
						// Not a valid file, but we'll catch that below.
					}
				}
			}
			if (characterSets.Count == 0) {
				characterSets['U'] = DEFAULT_CHARACTER_SETS[0];
				characterSets['L'] = DEFAULT_CHARACTER_SETS[1];
				characterSets['D'] = DEFAULT_CHARACTER_SETS[2];
			}
			if (templates.Count == 0) {
				templates.Add(DEFAULT_TEMPLATE);
			}
		}

		/// <summary>
		/// Generate a password that complies with AD complexity rules:
		/// minimum 8 characters (actually 10-14)
		/// Contains upper and lower-case letters, numbers and punctuation.
		/// </summary>
		/// <returns>A randomly-generated password</returns>
		public string GenerateRandomPassword() {
			Random rnd = new Random();
			// Pick the password template
			string template = templates[rnd.Next(templates.Count)];
			logger.WriteLine(LOG_SOURCE, "Selected template: {0}", template);
			// Now Pick a random character from each character set in order.
			int pwdLen = template.Length;
			char[] password = new char[pwdLen];
			int len = 0;
			for (int idx = 0; idx < pwdLen; idx++) {
				if (characterSets.ContainsKey(template[idx])) {
					string chars = characterSets[template[idx]];
					password[len++] = chars[rnd.Next(chars.Length)];
				} else {
					logger.WriteLine(LOG_SOURCE, "No such character set: {0}", template[idx]);
				}
			}
			return new string(password, 0, len);
		}
	}
}
