﻿using System;
using System.Diagnostics;
using System.EnterpriseServices.Internal;
using System.IO;
using System.Linq;
using System.Management;
using System.Security;
using System.Windows.Forms;
using System.Xml;
using com.swivelsecure.pinsafe.config.Properties;

namespace com.swivelsecure.pinsafe.config {

    /// <summary>
    /// Static utility methods for configuration.
    /// </summary>
    public class ConfigUtil {
        private const string IIS_RESET = "iisreset";

        public const string WEB_CONFIG = "\\web.config";
        private const string SOURCE_FOLDER = "web";
        private const string PINSAFE_CLIENT = "PINsafeClient";
        private const string PINSAFE_LOGIN = "PINsafeLogin";
        private const string PINSAFE_FILTER = "PINsafeSharePointFilter";
		private const string VERSION = "1.5.2.4";
        private const string PUBLIC_KEY = "7410363c6f9a59d4";
        private const string PINSAFE_CLIENT_ASSEMBLY_NAME = PINSAFE_CLIENT + ", Version=" + VERSION + ", Culture=neutral, PublicKeyToken=" + PUBLIC_KEY;
        private const string PINSAFE_LOGIN_ASSEMBLY_NAME = PINSAFE_LOGIN + ", Version=" + VERSION + ", Culture=neutral, PublicKeyToken=" + PUBLIC_KEY;
        private const string ASSEMBLY_FOLDER = VERSION + "__" + PUBLIC_KEY;
        private const string PINSAFE_NAMESPACE = "com.swivelsecure.pinsafe";
        private const string GAC_FOLDER = "\\assembly\\GAC_MSIL";

        // Only static methods.
        private ConfigUtil() {}


        public static void UpdateLocation(string newLocation, SharePointSettings sharePointSettings, PINsafeSettings pinsafeSettings, bool? requiresPINsafe) {
            CopyFiles(newLocation);
            RegisterFilter(newLocation + WEB_CONFIG, sharePointSettings, pinsafeSettings, requiresPINsafe);
        }

        public static void CopyFiles(string location) {
            // Copy all the files under the web folder to the specified location.
            if (!Directory.Exists(location)) {
                MessageBox.Show(String.Format(Resources.Location_Not_Found, location), Resources.Copy_Error, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            string source = GetWebSource();
            if (!Directory.Exists(source)) {
                MessageBox.Show(String.Format(Resources.Location_Not_Found, source), Resources.Copy_Error, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            CopyAllFiles(source, location);
        }

        private static string GetWebSource() {
            string source = Application.ExecutablePath;
            int idx = source.LastIndexOf('\\');
            source = ((idx <= 0) ? "" : source.Substring(0, idx + 1)) + SOURCE_FOLDER;
            return source;
        }

        private static void CopyAllFiles(string source, string location) {
            if (!Directory.Exists(source)) {
                MessageBox.Show(String.Format(Resources.Location_Not_Found, source), Resources.Copy_Error, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            if (!Directory.Exists(location)) {
                MessageBox.Show(String.Format(Resources.Location_Not_Found, location), Resources.Copy_Error, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            foreach (string subSource in Directory.GetDirectories(source)) {
                int idx = subSource.LastIndexOf('\\');
                string subLocation = location + (idx < 0 ? "\\" + subSource : subSource.Substring(idx));
                if (!Directory.Exists(subLocation)) {
                    Directory.CreateDirectory(subLocation);
                }
                CopyAllFiles(subSource, subLocation);
            }
            foreach (string file in Directory.GetFiles(source)) {
                int idx = file.LastIndexOf('\\');
                string locFile = location + (idx < 0 ? "\\" + file : file.Substring(idx));
                try {
                    File.Copy(file, locFile, true);
                } catch (Exception e) {
                    MessageBox.Show(e.Message, Resources.Error_Copying_File, MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        public static void StartIIS() {
            Process proc = new Process { StartInfo = new ProcessStartInfo(IIS_RESET) { Arguments = "/start" } };
            proc.Start();
            proc.WaitForExit();
        }

        public static void StopIIS() {
            Process proc = new Process { StartInfo = new ProcessStartInfo(IIS_RESET) { Arguments = "/stop" } };
            proc.Start();
            proc.WaitForExit();
        }

        public static void RestartIIS() {
            Process proc = Process.Start(IIS_RESET, "/noforce");
            if (proc != null) {
                proc.WaitForExit();
                MessageBox.Show(Resources.IIS_RESTARTED, Resources.HEADING_RESTART_IIS, MessageBoxButtons.OK);
            } else {
                MessageBox.Show(Resources.IIS_RESET_FAILED, Resources.HEADING_RESTART_IIS, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        public static void InstallGac(string location) {
            if (String.IsNullOrEmpty(location)) {
                return;
            }
            if (!location.EndsWith("\\")) {
                location += "\\";
            }
            try {
                Publish publish = new Publish();
                string fileName = location + PINSAFE_CLIENT + ".dll";
                publish.GacInstall(fileName);
                fileName = location + PINSAFE_FILTER + ".dll";
                publish.GacInstall(fileName);
                fileName = location + PINSAFE_LOGIN + ".dll";
                publish.GacInstall(fileName);
            } catch (SecurityException) {
                MessageBox.Show(Resources.Security_Error);
            }
        }

        public static void UninstallGac(string location) {
            if (String.IsNullOrEmpty(location)) {
                return;
            }
            if (!location.EndsWith("\\")) {
                location += "\\";
            }
            try {
                Publish publish = new Publish();
                string fileName = location + PINSAFE_FILTER + ".dll";
                publish.GacRemove(fileName);
                fileName = location + PINSAFE_CLIENT + ".dll";
                publish.GacRemove(fileName);
                fileName = location + PINSAFE_LOGIN + ".dll";
                publish.GacRemove(fileName);
            } catch (SecurityException) {
                MessageBox.Show(Resources.Security_Error);
            }
        }

        private static void UninstallOldGac(string assembly) {
            if (!Directory.Exists(assembly)) {
                return;
            }
            DirectoryInfo info = new DirectoryInfo(assembly);
            foreach (DirectoryInfo directory in
                info.GetDirectories().Where(directory => String.Compare(directory.Name, ASSEMBLY_FOLDER, StringComparison.OrdinalIgnoreCase) != 0)) {
                directory.Delete(true);
            }
        }

        private static void UninstallOldGac() {
            // Remove all PINsafe GAC assemblies.
            string sysFolder = Environment.GetFolderPath(Environment.SpecialFolder.System);
            int idx = sysFolder.LastIndexOf('\\');
            string gacFolder = sysFolder.Substring(0, idx) + GAC_FOLDER;
            if (!Directory.Exists(gacFolder)) {
                return;
            }
            UninstallOldGac(gacFolder + "\\" + PINSAFE_FILTER);
            UninstallOldGac(gacFolder + "\\" + PINSAFE_CLIENT);
            UninstallOldGac(gacFolder + "\\" + PINSAFE_LOGIN);
        }

        public static void UnregisterFilter(string location) {
            //
            if (!File.Exists(location)) {
                return;
            }
            bool changed = false;
            XmlDocument doc = new XmlDocument();
            try {
                // 
                doc.Load(location);
                XmlElement root = doc.DocumentElement;
                if (root != null) {
                    XmlElement level2;
                    XmlNodeList children;
                    XmlElement level1 = root["SharePoint"];
                    if (level1 != null) {
                        level2 = level1["SafeControls"];
                        if (level2 != null) {
                            children = level2.GetElementsByTagName("SafeControl");
                            XmlElement clientElem = null;
                            XmlElement loginElem = null;
                            foreach (XmlElement elem in children) {
                                string assemblyName = elem.GetAttribute("Assembly");
                                if (assemblyName.ToLower().StartsWith("pinsafeclient")) {
                                    // Found it
                                    clientElem = elem;
                                    changed = true;
                                    if (loginElem != null) {
                                        break;
                                    }
                                } else if (assemblyName.ToLower().StartsWith("pinsafelogin")) {
                                    // Found it
                                    loginElem = elem;
                                    changed = true;
                                    if (clientElem != null) {
                                        break;
                                    }
                                }
                            }
                            if (clientElem != null) {
                                level2.RemoveChild(clientElem);
                            }
                            if (loginElem != null) {
                                level2.RemoveChild(loginElem);
                            }
                        }
                    }
                    level1 = root["system.webServer"];
                    if (level1 != null) {
                        level2 = level1["modules"];
                        if (level2 != null) {
                            children = level2.GetElementsByTagName("add");
                            XmlElement modElem = null;
                            foreach (XmlElement elem in from XmlElement elem in children
                                                        let name = elem.GetAttribute("name")
                                                        where
                                                            String.Compare(name, "pinsafesharepointfilter",
                                                                           StringComparison.OrdinalIgnoreCase) == 0
                                                        select elem) {
                                modElem = elem;
                                changed = true;
                                break;
                            }
                            if (modElem != null) {
                                level2.RemoveChild(modElem);
                            }
                        }
                    }
                }
                if (changed) {
                    doc.Save(location);
                }
            } catch (Exception e) {
                MessageBox.Show(e.ToString(), Resources.Uninstall_Error, MessageBoxButtons.OK,
                                MessageBoxIcon.Error);
            }
        }

        public static void RegisterProvider(string location, SharePointSettings sharePointSettings) {
            try {
                XmlElement root;
                string errorMessage;
                XmlDocument doc;
                bool changed = ParseWebConfig(location, out doc, out root, out errorMessage);
                if (doc != null && root != null) {
                    changed |= RegisterProvider(doc, root, sharePointSettings);
                    if (changed && errorMessage == null) {
                        try {
                            doc.Save(location);
                        } catch (Exception e) {
                            MessageBox.Show(e.ToString(), String.Format(Resources.SAVE_FAILED, location), MessageBoxButtons.OK, MessageBoxIcon.Error);
                        }
                    }
                }
                if (errorMessage != null) {
                    MessageBox.Show(errorMessage, Resources.REGISTER_PROVIDER_ERROR, MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            } catch (Exception e) {
                MessageBox.Show(e.ToString(), Resources.REGISTER_PROVIDER_ERROR, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private static bool RegisterProvider(XmlDocument doc, XmlElement root, SharePointSettings sharePointSettings) {
            bool changed = false;
            XmlElement systemWebElem = root["system.web"];
            if (systemWebElem == null) {
                systemWebElem = doc.CreateElement("system.web");
                root.AppendChild(systemWebElem);
                changed = true;
            }
            XmlElement membershipElem = systemWebElem["membership"];
            if (membershipElem == null) {
                membershipElem = doc.CreateElement("membership");
                systemWebElem.AppendChild(membershipElem);
                changed = true;
            }
            XmlElement providersElem = membershipElem["providers"];
            if (providersElem == null) {
                providersElem = doc.CreateElement("providers");
                membershipElem.AppendChild(providersElem);
                changed = true;
            }
            XmlNodeList providersList = providersElem.GetElementsByTagName("add");
            bool found = false;
            XmlElement oldElem = null;
            string providerName = sharePointSettings.ProviderName;
            string oldProviderName = sharePointSettings.OldProviderName;
            bool hasOldProvider = !String.IsNullOrEmpty(oldProviderName);
            foreach (XmlElement providerElem in providersList) {
                string name = providerElem.GetAttribute("name");
                if (String.Compare(name, providerName, StringComparison.OrdinalIgnoreCase) == 0) {
                    found = true;
                } else if (hasOldProvider && String.Compare(name, oldProviderName, StringComparison.OrdinalIgnoreCase) == 0) {
                    oldElem = providerElem;
                }
                if (found && (oldElem != null || !hasOldProvider)) {
                    break;
                }
            }
            if (oldElem != null) {
                // Remove the old element.
                providersElem.RemoveChild(oldElem);
            }
            if (!found) {
                string providerType = Settings.Default.ProviderStrongName;
                string usernameAttr = sharePointSettings.UsernameAttr;
                XmlElement providerElem = doc.CreateElement("add");
                providerElem.SetAttribute("name", providerName);
                providerElem.SetAttribute("type", providerType);
                providerElem.SetAttribute("connectionStringName", sharePointSettings.ConnectionStringName);
                providerElem.SetAttribute("enableSearchMethods", "true");
                providerElem.SetAttribute("attributeMapUsername", usernameAttr);
                providersElem.AppendChild(providerElem);
                changed = true;
            }
            XmlElement connStringsElem = root["connectionStrings"];
            if (connStringsElem == null) {
                connStringsElem = doc.CreateElement("connectionStrings");
                root.InsertBefore(connStringsElem, systemWebElem);
                changed = true;
            }
            XmlNodeList stringsList = connStringsElem.GetElementsByTagName("add");
            string connectionStringName = sharePointSettings.ConnectionStringName;
            string oldConnectionStringName = sharePointSettings.OldConnectionStringName;
            bool hasOldConnectionStringName = !String.IsNullOrEmpty(oldConnectionStringName);
            XmlElement oldConnectionElem = null;
            XmlElement connectionElem = null;
            foreach (XmlElement stringElem in stringsList) {
                string name = stringElem.GetAttribute("name");
                if (String.Compare(name, connectionStringName, StringComparison.OrdinalIgnoreCase) == 0) {
                    connectionElem = stringElem;
                }
                if (hasOldConnectionStringName && String.Compare(oldConnectionStringName, name, StringComparison.OrdinalIgnoreCase) == 0) {
                    oldConnectionElem = stringElem;
                }
                if (connectionElem != null && (oldConnectionElem != null || !hasOldConnectionStringName)) {
                    break;
                }
            }
            if (oldConnectionElem != null) {
                connStringsElem.RemoveChild(oldConnectionElem);
            }
            if (connectionElem == null) {
                connectionElem = doc.CreateElement("add");
                connectionElem.SetAttribute("name", sharePointSettings.ConnectionStringName);
                connectionElem.SetAttribute("connectionString", sharePointSettings.ConnectionString);
                connStringsElem.AppendChild(connectionElem);
                changed = true;
            } else if (String.Compare(sharePointSettings.ConnectionString, connectionElem.GetAttribute("connectionString"), StringComparison.OrdinalIgnoreCase) != 0) {
                connectionElem.SetAttribute("connectionString", sharePointSettings.ConnectionString);
                changed = true;
            }
            return changed;
        }

        private static bool ParseWebConfig(string location, out XmlDocument doc, out XmlElement root, out string errorMessage) {
            doc = null;
            root = null;
            errorMessage = null;
            if (File.Exists(location)) {
                bool changed = false;
                doc = new XmlDocument();
                // 
                try {
                    doc.Load(location);
                } catch (XmlException) {
                    errorMessage = "\"" + location + "\" is not a valid XML file";
                    return false;
                }
                root = doc.DocumentElement;
                if (root == null) {
                    root = doc.CreateElement("configuration");
                    doc.AppendChild(root);
                    changed = true;
                }
                return changed;
            }
            // The file doesn't exist.
            errorMessage = "The file \"" + location + "\" does not exist";
            return false;
        }

        public static void RegisterFilter(string location, SharePointSettings sharePointSettings, PINsafeSettings pinsafeSettings, bool? requiresPINsafe) {

            // Set up the basic structure
            try {
                string errorMessage;
                XmlDocument doc;
                XmlElement root;
                bool changed = ParseWebConfig(location, out doc, out root, out errorMessage);
                if (errorMessage != null) {
                    MessageBox.Show(errorMessage, Resources.REGISTER_FILTER_ERROR, MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }
                changed |= RegisterProvider(doc, root, sharePointSettings);

                // Register PINsafeClient as a Sharepoint safe control
                XmlElement level1 = root["SharePoint"];
                if (level1 == null) {
                    level1 = doc.CreateElement("SharePoint");
                    root.AppendChild(level1);
                    changed = true;
                }
                XmlElement level2 = level1["SafeControls"];
                if (level2 == null) {
                    level2 = doc.CreateElement("SafeControls");
                    level1.AppendChild(level2);
                    changed = true;
                }
                XmlNodeList children = level2.GetElementsByTagName("SafeControl");
                bool clientInstalled = false;
                bool loginInstalled = false;
                foreach (XmlElement elem in children) {
                    string assemblyName = elem.GetAttribute("Assembly");
                    if (assemblyName.ToLower().StartsWith("pinsafeclient")) {
                        clientInstalled = true;
                    } else if (assemblyName.ToLower().StartsWith("pinsafelogin")) {
                        loginInstalled = true;
                    }
                    if (clientInstalled && loginInstalled) {
                        break;
                    }
                }
                XmlElement level3;
                if (!clientInstalled) {
                    level3 = doc.CreateElement("SafeControl");
                    level3.SetAttribute("Assembly", PINSAFE_CLIENT_ASSEMBLY_NAME);
                    level3.SetAttribute("Namespace", PINSAFE_NAMESPACE);
                    level3.SetAttribute("TypeName", "*");
                    level3.SetAttribute("Safe", "True");
                    level2.AppendChild(level3);
                    changed = true;
                }
                if (!loginInstalled) {
                    level3 = doc.CreateElement("SafeControl");
                    level3.SetAttribute("Assembly", PINSAFE_LOGIN_ASSEMBLY_NAME);
                    level3.SetAttribute("Namespace", PINSAFE_NAMESPACE);
                    level3.SetAttribute("TypeName", "*");
                    level3.SetAttribute("Safe", "True");
                    level2.AppendChild(level3);
                    changed = true;
                }

                // Ensure the filter is installed as a HTTP module
                level1 = root["system.webServer"];
                if (level1 == null) {
                    level1 = doc.CreateElement("system.webServer");
                    root.AppendChild(level1);
                    changed = true;
                }
                level2 = level1["modules"];
                if (level2 == null) {
                    level2 = doc.CreateElement("modules");
                    level1.AppendChild(level2);
                    changed = true;
                }
                children = level2.GetElementsByTagName("add");
                bool installed = false;
                XmlElement sibling = null;
                foreach (XmlElement elem in children) {
                    string name = elem.GetAttribute("name");
                    if (String.Compare(name, "pinsafesharepointfilter", StringComparison.OrdinalIgnoreCase) == 0) {
                        installed = true;
                        break;
                    }
                    if (String.Compare(name, "FederatedAuthentication", StringComparison.OrdinalIgnoreCase) == 0) {
                        sibling = elem;
                    }
                }
                if (!installed) {
                    level3 = doc.CreateElement("add");
                    level3.SetAttribute("name", "PINsafeSharepointFilter");
                    level3.SetAttribute("type", "com.swivelsecure.Sharepoint.SharepointFilter");
                    if (sibling == null) {
                        level2.AppendChild(level3);
                    } else {
                        level2.InsertBefore(level3, sibling);
                    }
                    changed = true;
                }
                level1 = root["appSettings"];
                if (level1 == null) {
                    level1 = doc.CreateElement("appSettings");
                    root.AppendChild(level1);
                    changed = true;
                }
                if (changed) {
                    try {
                        doc.Save(location);
                    } catch {
                        MessageBox.Show(String.Format(Resources.SAVE_FAILED, location), Resources.ERROR, MessageBoxButtons.OK, MessageBoxIcon.Error);
                        return;
                    }
                }
                // Make sure the settings are initialised.
                if (!pinsafeSettings.SaveSettings(location, requiresPINsafe)) {
                    MessageBox.Show(String.Format(Resources.SAVE_FAILED, location), Resources.ERROR, MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            } catch (Exception e) {
                MessageBox.Show(e.ToString(), Resources.REGISTER_FILTER_ERROR, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        public static void CreateShare() {
            // Create a network share for the configuration store.
            string sharePath = PINsafeSettings.CommonDataPath;
            if (sharePath.EndsWith("\\")) {
                sharePath = sharePath.Substring(0, sharePath.Length - 1); // Remove the trailing backslash.
            }
            ManagementClass shareMgmt = new ManagementClass("Win32_Share");
            // Check if the share already exists.
            ManagementObjectSearcher searcher = new ManagementObjectSearcher(new SelectQuery("Win32_Share", PINsafeSettings.SHARE_SEARCH_PATH));
            if (searcher.Get().Count == 0) {
                ManagementBaseObject inParams = shareMgmt.GetMethodParameters("Create");
                inParams["Name"] = PINsafeSettings.SHARE_NAME;
                inParams["Path"] = sharePath;
                inParams["Type"] = 0;
                ManagementBaseObject outParams = shareMgmt.InvokeMethod("Create", inParams, null);
                if (outParams == null || ((uint) (outParams.Properties["ReturnValue"].Value)) != 0) {
                    MessageBox.Show(Resources.Share_Config_Failed, Resources.Share_Config_Title, MessageBoxButtons.OK, MessageBoxIcon.Error);
                } else {
                    MessageBox.Show(
                        Resources.Share_Config_Succeeded,
                        Resources.Share_Config_Title, MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
            } else {
                MessageBox.Show(Resources.Share_Config_Already, Resources.Share_Config_Title, MessageBoxButtons.OK,
                                MessageBoxIcon.Warning);
            }
        }

        public static void RemoveShare() {
            // Check if the share already exists.
            ManagementObjectSearcher searcher = new ManagementObjectSearcher(new SelectQuery("Win32_Share", PINsafeSettings.SHARE_SEARCH_PATH));
            foreach (ManagementObject item in searcher.Get()) {
                item.Delete();
            }
        }

        /// <summary>
        /// Remove all PINsafe files from the specified location.
        /// </summary>
        /// <param name="location"></param>
        public static void RemoveFiles(string location) {
            string centralLoc = GetWebSource();
            RemoveAllFiles(centralLoc, location);
        }

        /// <summary>
        /// Remove all files from the second location that exist in the first.
        /// </summary>
        /// <param name="source">The central location (not altered)</param>
        /// <param name="location">The location from which to remove files.</param>
        private static void RemoveAllFiles(string source, string location) {
            if (!Directory.Exists(source)) {
                return;
            }
            if (!Directory.Exists(location)) {
                return;
            }
            foreach (string subSource in Directory.GetDirectories(source)) {
                int idx = subSource.LastIndexOf('\\');
                string subLocation = location + (idx < 0 ? "\\" + subSource : subSource.Substring(idx));
                if (Directory.Exists(subLocation)) {
                    RemoveAllFiles(subSource, subLocation);
                }
            }
            foreach (string locFile in from file in Directory.GetFiles(source)
                                       let idx = file.LastIndexOf('\\')
                                       select location + (idx < 0 ? "\\" + file : file.Substring(idx))) {
                if (!File.Exists(locFile)) {
                    continue;
                }
                try {
                    File.Delete(locFile);
                } catch {
                    MessageBox.Show(String.Format(Resources.Delete_Failed_Message, locFile), Resources.Delete_Failed_Title, MessageBoxButtons.OK,
                                    MessageBoxIcon.Warning);
                }
            }
            if (!Directory.GetDirectories(location).Any() && !Directory.GetFiles(location).Any()) {
                Directory.Delete(location);
            }
        }

        /// <summary>
        /// Upgrade from an existing installation (version 1.2 or later).
        /// </summary>
        public static void Upgrade(string location) {
            try {
                Cursor.Current = Cursors.WaitCursor;
                StopIIS();
                // Unregister the old DLLs from GAC.
                UninstallOldGac();
                InstallGac(location);
                SharePointSettings settings = new SharePointSettings();
                PINsafeSettings pinsafeSettings = PINsafeSettings.SettingsFromStorage;
                foreach (SharePointSettings.SharePointApplication app in settings.GetLocations()) {
                    UpdateLocation(app.Location, settings, pinsafeSettings, app.RequiresPINsafe);
                }
            } finally {
                StartIIS();
                Cursor.Current = Cursors.Default;
            }
        }
    }
}
