﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Xml;
using com.swivelsecure.pinsafe.config.Properties;
using Microsoft.Win32;

namespace com.swivelsecure.pinsafe.config {
    public class SharePointSettings {

        public class SharePointApplication {

            public string Location { get; set; }

            public bool RequiresPINsafe { get; set; }

            private const string REQUIRES_PINSAFE = " (Two Factor)";
            private const string NO_PINSAFE = " (One Factor)";

            private string AppType {
                get { return (RequiresPINsafe ? REQUIRES_PINSAFE : NO_PINSAFE); }
            }

            public SharePointApplication() : this("", true) {}

            public SharePointApplication(string location) : this(location, true) {}

            public SharePointApplication(string location, bool requirePS) {
                Location = location;
                RequiresPINsafe = requirePS;
            }

            public override String ToString() {
                return Location + AppType;
            }

            public String Display {
                get { return ToString(); }
            }
        }

        private const string PROVIDER_NAME_ELEM = "ProviderName";
        private const string CONNECTION_STRING_NAME_ELEM = "ConnectionStringName";
        private const string CONNECTION_STRING_ELEM = "ConnectionString";
        private const string USERNAME_ATTR_ELEM = "UsernameAttr";
        private const string CENTRAL_ADMIN_PATH_ELEM = "CentralAdminPath";
        private const string SECURITY_TOKEN_PATH_ELEM = "SecurityTokenPath";
        private const string STORAGE_FILE = "SharePointSettings.data";
        private const string DEFAULT_PROVIDER_NAME = "ADMembership";
        private const string DEFAULT_CONNECTION_STRING_NAME = "ADConnection";
        private const string DEFAULT_CONNECTION_STRING = "LDAP://server/dc=domain,dc=local";
        private const string DEFAULT_USERNAME_ATTR = "sAMAccountName";
        private const string EXISTS = "exists";
        private const string APP_CONTAINER_ELEM = "Applications";
        private const string APPLICATION_ELEM = "Application";
        private const string CENTRAL_SERVER_ELEM = "ConfigServer";
        private const string PINSAFE_ATTR = "pinsafe";

        public string ProviderName { get; set; }
        public string ConnectionStringName { get; set; }
        public string ConnectionString { get; set; }
        public string UsernameAttr { get; set; }
        public string ConfigServer { get; set; }
        public string OldConnectionStringName { get; set; }
        public string OldProviderName { get; set; }

        private string centralAdminPath;
        public string CentralAdminPath {
            get { return centralAdminPath; }
            set { centralAdminPath = value; }
        }

        private string securityTokenPath;
        public string SecurityTokenPath {
            get { return securityTokenPath; }
            set { securityTokenPath = value; }
        }
        public bool CentralAdminExists { get; set; }
        public bool SecurityTokenExists { get; set; }

        private readonly List<SharePointApplication> appLocations = new List<SharePointApplication>();

        public bool AddLocation(string location, bool requirePS) {
            foreach (
                SharePointApplication loc in
                    appLocations.Where(
                        loc => String.Compare(loc.Location, location, StringComparison.OrdinalIgnoreCase) == 0)) {
                if (loc.RequiresPINsafe != requirePS) {
                    loc.RequiresPINsafe = requirePS;
                    return true;
                }
                return false;
            }
            appLocations.Add(new SharePointApplication(location, requirePS));
            return true;
        }

        public bool RemoveLocation(string location) {
            int oldIdx = -1;
            for (int idx = 0; idx<appLocations.Count; idx++) {
                SharePointApplication loc = appLocations[idx];
                if (String.Compare(loc.Location, location, StringComparison.OrdinalIgnoreCase) != 0) {
                    continue;
                }
                oldIdx = idx;
                break;
            }
            if (oldIdx >= 0) {
                appLocations.RemoveAt(oldIdx);
                return true;
            }
            return false;
        }

        public ReadOnlyCollection<SharePointApplication> GetLocations() {
            return appLocations.AsReadOnly();
        }

        public SharePointSettings() {
            ReadFromStorage();
        }

        public void ReadFromStorage() {
            if (!PINsafeSettings.ConfigFileExists(STORAGE_FILE)) {
                // config file doesn't exist - use defaults.
                SetDefaults();
                return;
            }
            XmlDocument configDoc = new XmlDocument();
            try {
                using (Stream stream = PINsafeSettings.OpenConfigFileForRead(STORAGE_FILE)) {
                    try {
                        configDoc.Load(stream);
                    } catch {
                        SetDefaults();
                        return;
                    }
                }
            } catch {
                SetDefaults();
                return;
            }
            XmlElement settingsElem = configDoc.DocumentElement;
            if (settingsElem == null) {
                SetDefaults();
                return;
            }
            ProviderName = GetConfigString(settingsElem, PROVIDER_NAME_ELEM, DEFAULT_PROVIDER_NAME);
            ConnectionStringName = GetConfigString(settingsElem, CONNECTION_STRING_NAME_ELEM, DEFAULT_CONNECTION_STRING_NAME);
            ConnectionString = GetConfigString(settingsElem, CONNECTION_STRING_ELEM, DEFAULT_CONNECTION_STRING);
            UsernameAttr = GetConfigString(settingsElem, USERNAME_ATTR_ELEM, DEFAULT_USERNAME_ATTR);
            ConfigServer = GetConfigString(settingsElem, CENTRAL_SERVER_ELEM, "");
            CentralAdminExists = GetConfigStringAndBool(settingsElem, CENTRAL_ADMIN_PATH_ELEM, "", out centralAdminPath);
            SecurityTokenExists = GetConfigStringAndBool(settingsElem, SECURITY_TOKEN_PATH_ELEM, "", out securityTokenPath);
            if ("".Equals(CentralAdminPath) && "".Equals(SecurityTokenPath)) {
                SetDefaultPaths();
            }
            appLocations.Clear();
            XmlElement locElem = settingsElem[APP_CONTAINER_ELEM];
            if (locElem == null) {
                return;
            }
            XmlNodeList locElems = locElem.GetElementsByTagName(APPLICATION_ELEM);
            foreach (XmlNode node in locElems) {
                if (node is XmlElement) {
                    AddLocation(node as XmlElement);
                }
            }
        }

        private void AddLocation(XmlElement elem) {
            bool requiresPS = true;
            if (elem.HasAttribute(PINSAFE_ATTR)) {
                if (!Boolean.TryParse(elem.GetAttribute(PINSAFE_ATTR), out requiresPS)) {
                    requiresPS = true;
                }
            }
            appLocations.Add(new SharePointApplication(elem.InnerText, requiresPS));
        }

        public bool ReadFromSharedConfig() {
            if (String.IsNullOrEmpty(ConfigServer)) {
                return false;
            }
            if (!PINsafeSettings.SharedConfigFileExists(ConfigServer, STORAGE_FILE)) {
                return false;
            }
            XmlDocument configDoc = new XmlDocument();
            try {
                using (Stream stream = PINsafeSettings.OpenSharedConfig(ConfigServer, STORAGE_FILE)) {
                    try {
                        configDoc.Load(stream);
                    } catch {
                        return false;
                    }
                }
            } catch {
                return false;
            }
            XmlElement settingsElem = configDoc.DocumentElement;
            if (settingsElem == null) {
                return false;
            }
            ProviderName = GetConfigString(settingsElem, PROVIDER_NAME_ELEM, DEFAULT_PROVIDER_NAME);
            ConnectionStringName = GetConfigString(settingsElem, CONNECTION_STRING_NAME_ELEM, DEFAULT_CONNECTION_STRING_NAME);
            ConnectionString = GetConfigString(settingsElem, CONNECTION_STRING_ELEM, DEFAULT_CONNECTION_STRING);
            UsernameAttr = GetConfigString(settingsElem, USERNAME_ATTR_ELEM, DEFAULT_USERNAME_ATTR);
            CentralAdminPath = GetConfigString(settingsElem, CENTRAL_ADMIN_PATH_ELEM, "");
            CentralAdminExists = Directory.Exists(CentralAdminPath);
            SecurityTokenPath = GetConfigString(settingsElem, SECURITY_TOKEN_PATH_ELEM, "");
            SecurityTokenExists = Directory.Exists(SecurityTokenPath);
            if ("".Equals(CentralAdminPath) && "".Equals(SecurityTokenPath)) {
                SetDefaultPaths();
            }
            appLocations.Clear();
            XmlElement locElem = settingsElem[APP_CONTAINER_ELEM];
            if (locElem == null) {
                return true;
            }
            XmlNodeList locElems = locElem.GetElementsByTagName(APPLICATION_ELEM);
            foreach (XmlNode elem in locElems) {
                if (elem is XmlElement) {
                    AddLocation(elem as XmlElement);
                }
            }
            return true;
        }

        private void SetDefaults() {
            ProviderName = DEFAULT_PROVIDER_NAME;
            ConnectionStringName = DEFAULT_CONNECTION_STRING_NAME;
            ConnectionString = DEFAULT_CONNECTION_STRING;
            UsernameAttr = DEFAULT_USERNAME_ATTR;
            ConfigServer = "";
            appLocations.Clear();
            SetDefaultPaths();
            WriteToStorage();
        }

        private static string GetConfigString(XmlElement settingsElem, string name, string defVal) {
            XmlElement elem = settingsElem[name];
            return (elem == null) ? defVal : elem.InnerText;
        }

        private static bool GetConfigStringAndBool(XmlElement settingsElem, string name, string defVal, out string value) {
            XmlElement elem = settingsElem[name];
            if (elem == null) {
                value = defVal;
                return false;
            }
            value = elem.InnerText;
            bool result;
            if (!Boolean.TryParse(elem.GetAttribute(EXISTS), out result)) {
                result = false;
            }
            return result;
        }

        public bool WriteToStorage() {
            XmlDocument configDoc = new XmlDocument();
            XmlElement settingsElem = configDoc.CreateElement("SharePointSettings");
            configDoc.AppendChild(settingsElem);
            SetConfigString(configDoc, settingsElem, PROVIDER_NAME_ELEM, ProviderName);
            SetConfigString(configDoc, settingsElem, CONNECTION_STRING_NAME_ELEM, ConnectionStringName);
            SetConfigString(configDoc, settingsElem, CONNECTION_STRING_ELEM, ConnectionString);
            SetConfigString(configDoc, settingsElem, USERNAME_ATTR_ELEM, UsernameAttr);
            SetConfigString(configDoc, settingsElem, CENTRAL_SERVER_ELEM, ConfigServer);
            SetConfigStringAndBool(configDoc, settingsElem, CENTRAL_ADMIN_PATH_ELEM, CentralAdminPath, CentralAdminExists);
            SetConfigStringAndBool(configDoc, settingsElem, SECURITY_TOKEN_PATH_ELEM, SecurityTokenPath, SecurityTokenExists);
            XmlElement appContainer = configDoc.CreateElement(APP_CONTAINER_ELEM);
            foreach (SharePointApplication app in appLocations) {
                XmlElement elem = configDoc.CreateElement(APPLICATION_ELEM);
                elem.InnerText = app.Location;
                elem.SetAttribute(PINSAFE_ATTR, app.RequiresPINsafe.ToString(CultureInfo.InvariantCulture));
                appContainer.AppendChild(elem);
            }
            settingsElem.AppendChild(appContainer);
            try {
                using (Stream stream = PINsafeSettings.OpenConfigFileForWrite(STORAGE_FILE)) {
                    configDoc.Save(stream);
                }
            } catch {
                return false;
            }
            return true;
        }

        private static void SetConfigStringAndBool(XmlDocument configDoc, XmlElement settingsElem, string name, string value, bool exists) {
            XmlElement elem = SetConfigString(configDoc, settingsElem, name, value);
            elem.SetAttribute(EXISTS, exists.ToString(CultureInfo.InvariantCulture));
        }

        private static XmlElement SetConfigString(XmlDocument configDoc, XmlElement settingsElem, string name, string value) {
            XmlElement elem = configDoc.CreateElement(name);
            elem.InnerText = value;
            settingsElem.AppendChild(elem);
            return elem;
        }

        private void SetDefaultPaths() {
            string centLoc = Settings.Default.DefaultSharePointPath;
            using (RegistryKey key = Registry.LocalMachine.OpenSubKey(@"Software\Microsoft\Shared Tools\Web Server Extensions\14.0\wss")) {
                if (key != null) {
                    object caUrl = key.GetValue("CentralAdministrationUrl");
                    if (caUrl != null) {
                        centLoc += caUrl.ToString().Substring(caUrl.ToString().LastIndexOf(':') + 1);
                    }
                }
            }
            CentralAdminPath = centLoc;
            CentralAdminExists = true;
            string stLoc = Settings.Default.DefaultSecurityTokenPath;
            using (RegistryKey key = Registry.LocalMachine.OpenSubKey(@"Software\Microsoft\Shared Tools\Web Server Extensions\14.0")) {
                if (key != null) {
                    object stPath = key.GetValue("Location");
                    if (stPath != null) {
                        stLoc = stPath + @"WebServices\SecurityToken";
                    }
                }
            }
            SecurityTokenPath = stLoc;
            SecurityTokenExists = true;
        }
    }
}
