﻿using System;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using com.swivelsecure.pinsafe.config.Properties;

namespace com.swivelsecure.pinsafe.config {
    public partial class UpgradeSettings : Form {

        private const string CONFIG_FILE = "\\web.config";
        private const string SWIVEL_APP_FOLDER = "\\Swivel_Secure_Ltd";
        private const string USER_CONFIG_FILE = "\\1.0.0.0\\user.config";
        private const string OLD_FORMS_LOC = "_forms";
        private const string NEW_FORMS_LOC = "pinsafe";
        const string OLD_FORMS_URL = "/" + OLD_FORMS_LOC + "/";
        const string NEW_FORMS_URL = "/" + NEW_FORMS_LOC + "/";
        private const string BIN_FOLDER = "\\bin\\";

        private static readonly string[] OLD_FILE_NAMES = {
                                                     "ConfigApp.exe",
                                                     "RegisterFilter.exe",
                                                     "RegisterFilter.exe.config",
                                                     "Swivel.ico"
                                                 };

        private const string START_MENU_NAME = "\\PINsafe Filter for Sharepoint 2010\\";
        private const string COMMON_START_MENU_NAME = "\\Programs" + START_MENU_NAME;

        private static readonly string[] OLD_SHORTCUTS = {
                                                     "Register Filter.lnk",
                                                     "Unregister Filter.lnk"
                                                 };

        public UpgradeSettings() {
            InitializeComponent();
        }

        private void BtnCancelClick(object sender, EventArgs e) {
            if (MessageBox.Show(Resources.Cancel_Import_Message, Resources.Cancel_Import_Title, MessageBoxButtons.YesNo, MessageBoxIcon.Warning) == DialogResult.No) {
                return;
            }
            DialogResult = DialogResult.Cancel;
            Close();
        }

        private void BtnSelectClick(object sender, EventArgs e) {
            FolderBrowserDialog dlg = new FolderBrowserDialog {Description = Resources.Select_Application_Folder, SelectedPath = lblImportFolder.Text, ShowNewFolderButton = false};
            if (dlg.ShowDialog() == DialogResult.Cancel) {
                return;
            }
            lblImportFolder.Text = dlg.SelectedPath;
        }

        private void BtnImportClick(object sender, EventArgs e) {
            string folder = lblImportFolder.Text;
            if (String.IsNullOrEmpty(folder)) {
                MessageBox.Show(String.Format(Resources.No_Folder_Message, folder), Resources.No_Folder_Title, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            if (!Directory.Exists(folder)) {
                MessageBox.Show(String.Format(Resources.No_Such_Folder_Message, folder), Resources.No_Such_Folder_Title, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            folder = (folder.EndsWith("\\") ? folder.Substring(0, folder.Length - 1) : folder);
            string configFile = folder + CONFIG_FILE;
            if (!File.Exists(configFile)) {
                MessageBox.Show(String.Format(Resources.Conf_File_Not_Found_Message, folder), Resources.Conf_File_Not_Found_Title, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            PINsafeSettings settings = PINsafeSettings.SettingsFromLocation(configFile);
            if (settings == null) {
                MessageBox.Show(String.Format(Resources.Conf_File_Error_Message, folder), Resources.Conf_File_Error_Title, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            // We need to stop IIS while we're modifying files etc.
            ConfigUtil.StopIIS();
            // Uninstall the old assemblies from GAC.
            string binFolder = folder + BIN_FOLDER;
            ConfigUtil.UninstallGac(binFolder);
            // Unregister the old filter.
            ConfigUtil.UnregisterFilter(configFile);
            // Remove the old PINsafe settings.
            PINsafeSettings.RemoveSettings(configFile);
            // We need to move the location of the login pages.
            settings.LogonUrl = settings.LogonUrl.Replace(OLD_FORMS_URL, NEW_FORMS_URL);
            settings.LogoffUrl = settings.LogoffUrl.Replace(OLD_FORMS_URL, NEW_FORMS_URL);
            string[] oldPaths = settings.ExcludedPaths;
            string[] newPaths = new string[oldPaths.Length];
            for (int idx = 0; idx < oldPaths.Length; idx++) {
                newPaths[idx] = oldPaths[idx].Replace(OLD_FORMS_URL, NEW_FORMS_URL);
            }
            settings.ExcludedPaths = newPaths;
            // Save the new settings to storage.
            settings.WriteSettingsToStorage();
            // Rename the old login path folder.);
            if (Directory.Exists(folder + "\\" + OLD_FORMS_LOC)) {
                Directory.Move(folder + "\\" + OLD_FORMS_LOC, folder + "\\" + NEW_FORMS_LOC);
            }

            // OK, let's see if we can find the old registration settings.
            bool checkOldSettings = CheckOldSettings(folder);
            // Now deploy the new configuration to the existing location.
            if (checkOldSettings) {
                RemoveOldFiles(binFolder);
                ConfigUtil.CopyFiles(folder);
                ConfigUtil.RegisterFilter(configFile, new SharePointSettings(), settings, true);
            }
            ConfigUtil.StartIIS();
            if (checkOldSettings) {
                MessageBox.Show(Resources.Settings_Upgraded_Message, Resources.Settings_Upgraded_Title,
                                MessageBoxButtons.OK, MessageBoxIcon.Information);
            } else {
                MessageBox.Show(Resources.Settings_Partially_Upgraded_Message,
                                Resources.Settings_Partially_Upgraded_Title, MessageBoxButtons.OK,
                                MessageBoxIcon.Warning);
            }
            DialogResult = DialogResult.OK;
            Close();
        }

        [DllImport("shell32.dll")]
        static extern bool SHGetSpecialFolderPath(IntPtr hwndOwner, [Out] StringBuilder lpszPath, int nFolder, bool fCreate);

        private const int CSIDL_COMMON_STARTMENU = 0x16;

        private static void RemoveOldFiles(string location) {
            foreach (string name in OLD_FILE_NAMES) {
                File.Delete(location + name);
            }
            string folderName = Environment.GetFolderPath(Environment.SpecialFolder.StartMenu);
            if (Directory.Exists(folderName + START_MENU_NAME)) {
                folderName = folderName + START_MENU_NAME;
                foreach (string name in OLD_SHORTCUTS) {
                    string fullName = folderName + name;
                    if (File.Exists(fullName)) {
                        File.Delete(fullName);
                    }
                }
            }
            StringBuilder buffer = new StringBuilder(260);
            if (!SHGetSpecialFolderPath(IntPtr.Zero, buffer, CSIDL_COMMON_STARTMENU, false)) {
                return;
            }
            if (Directory.Exists(buffer + COMMON_START_MENU_NAME)) {
                folderName = buffer + COMMON_START_MENU_NAME;
                foreach (string name in OLD_SHORTCUTS) {
                    string fullName = folderName + name;
                    if (File.Exists(fullName)) {
                        File.Delete(fullName);
                    }
                }
            }
        }

        private static bool CheckOldSettings(string appPath) {
            bool loaded = false;
            string settingsFolder = Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData) + SWIVEL_APP_FOLDER;
            if (Directory.Exists(settingsFolder)) {
                string[] subFolders = Directory.GetDirectories(settingsFolder, "RegisterFilter.exe_*");
                if (subFolders.Length == 1) {
                    string settingsFile = subFolders[0] + USER_CONFIG_FILE;
                    if (File.Exists(settingsFile)) {
                        try {
                            XmlDocument doc = new XmlDocument();
                            using (Stream stream = new FileStream(settingsFile, FileMode.Open)) {
                                doc.Load(stream);
                            }
                            if (doc.DocumentElement != null) {
                                XmlElement settingsElem = doc.DocumentElement["userSettings"];
                                if (settingsElem != null) {
                                    settingsElem = settingsElem["RegisterFilter.Properties.Settings"];
                                    if (settingsElem != null) {
                                        XmlNodeList settingElems = settingsElem.GetElementsByTagName("setting");
                                        SharePointSettings spSettings = new SharePointSettings();
                                        spSettings.AddLocation(appPath, true);
                                        foreach (XmlElement settingElem in settingElems) {
                                            string name = settingElem.GetAttribute("name");
                                            XmlElement valElem = settingElem["value"];
                                            if (valElem != null) {
                                                string value = valElem.InnerText;
                                                if (name.Equals("ProviderName")) {
                                                    spSettings.ProviderName = value;
                                                } else if (name.Equals("ConnectionStringName")) {
                                                    spSettings.ConnectionStringName = value;
                                                } else if (name.Equals("ConnectionString")) {
                                                    spSettings.ConnectionString = value;
                                                } else if (name.Equals("CentralAdminPath")) {
                                                    spSettings.CentralAdminPath = value;
                                                    spSettings.CentralAdminExists = true;
                                                } else if (name.Equals("SecurityTokenPath")) {
                                                    spSettings.SecurityTokenPath = value;
                                                    spSettings.SecurityTokenExists = true;
                                                } else if (name.Equals("UsernameAttr")) {
                                                    spSettings.UsernameAttr = value;
                                                }
                                            }
                                        }
                                        spSettings.WriteToStorage();
                                        loaded = true;
                                    }
                                }
                            }
                        } catch {
                            // OK, if it fails, just ignore it.
                        }
                    }
                }
            }
            return loaded;
        }
    }
}
