﻿using System;
using System.Linq;
using System.Text;
using System.Web;
using com.swivelsecure.pinsafe;

namespace com.swivelsecure.Sharepoint {
    internal class SharepointFilter : IHttpModule {
        private const string COOKIE_NAME = "PINSAFE";
        private const string CP_COOKIE_NAME = "PINSAFECP";
        public const string CPWD_COOKIE_NAME = "CHANGEPWD";
        private const string FEDAUTH_COOKIE_NAME = "FedAuth";
	    private const string LOG_SOURCE = "FILTER";
        private PINsafeSettings settings;

        public void Init(HttpApplication context) {
            settings = PINsafeSettings.SettingsFromConfiguration;
            context.BeginRequest += BeginRequest;
            Debug.WriteLine(LOG_SOURCE, "Swivel filter initialised");
        }

        private DebugLogger Debug {
            get { return settings.Logger; }
        }

        private void BeginRequest(object sender, EventArgs e) {
            if (!settings.FilterEnabled) {
                return;
            }
            HttpApplication app = (HttpApplication) sender;
            HttpContext context = app.Context;
            HttpRequest request = context.Request;
            string logonUrl = settings.LogonUrl;
            // Check whether this request requires authentication.
            string requestUrl = request.Path;
            if (!CheckAuthenticationRequired(requestUrl)) {
//                Debug.WriteLine(LOG_SOURCE, "Swivel authentication not required for \"{0}\"", requestUrl);
                return;
            }
//            Debug.WriteLine(LOG_SOURCE, "Checking requested URL \"{0}\"", requestUrl);

            // Authentication is required, check if the PINsafe cookie exists and is valid.
            HttpCookie pinsafeCookie = request.Cookies[COOKIE_NAME];
            bool cookieValid = false;
            bool passwordChangeRequired = false;
            PINsafeClient client = new PINsafeClient(settings);
            HttpCookie cpwdCookie = request.Cookies[CPWD_COOKIE_NAME];
            if (cpwdCookie != null && !String.IsNullOrEmpty(cpwdCookie.Value)) {
                // Password change required.
                passwordChangeRequired = true;
            }
            if (pinsafeCookie != null) {
                bool? requiresPINsafe = settings.RequiresPINsafe;
                cookieValid = (requiresPINsafe != null && !(bool)requiresPINsafe) ? client.SingleFactorCookieValid(pinsafeCookie.Value) : client.TwoFactorCookieValid(pinsafeCookie.Value);
                if (cookieValid) {
                    // Check if we have the FedAuth cookie.
                    HttpCookie fedAuthCookie = request.Cookies[FEDAUTH_COOKIE_NAME];
                    if (fedAuthCookie == null) {
                        if (passwordChangeRequired) {
                            cookieValid = false;
                        } else {
                            // We've logged out of Sharepoint, so log out of PINsafe.
                            pinsafeCookie.Expires = new DateTime(1970, 1, 1);
                            context.Response.Cookies.Remove(COOKIE_NAME);
                            context.Response.Cookies.Add(pinsafeCookie);
                            cookieValid = false;
                        }
                    }
                }
            }
            // If not, redirect to the logon page.
            if (!cookieValid) {
                StringBuilder url = new StringBuilder();
                if (passwordChangeRequired) {
                    url.Append(settings.ChangePasswordUrl);
//                    Debug.WriteLine(LOG_SOURCE, "Redirect to change password page");
                } else {
                    // Check for change PIN cookie.
                    HttpCookie cpCookie = request.Cookies[CP_COOKIE_NAME];
                    bool cpCookieValid = false;
                    if (cpCookie != null) {
                        cpCookieValid = client.CPCookieValid(cpCookie.Value);
                        if (cpCookieValid) {
                            // Check if we have the FedAuth cookie.
                            HttpCookie fedAuthCookie = request.Cookies[FEDAUTH_COOKIE_NAME];
                            if (fedAuthCookie == null) {
                                cpCookieValid = false;
                            }
                        }
                    }
                    url.Append(cpCookieValid ? settings.ChangePinUrl : logonUrl);
//                    Debug.WriteLine(LOG_SOURCE, cpCookieValid ? "Redirect to change PIN page" : "No valid cookie - redirect to logon page");
                }
                url.Append("?Source=");
                url.Append(requestUrl);
                url.Append("&mobile=0");
                foreach (string arg in request.QueryString) {
                    url.Append("&");
                    url.Append(arg);
                    url.Append("=");
                    url.Append(HttpUtility.UrlEncode(request.Params[arg]));
                }
                context.Response.Redirect(url.ToString());
            } else {
                // If it is valid, allow request to continue.
                // Update the PINsafe cookie expiry time if necessary
                string newCookieValue = client.UpdateCookie(pinsafeCookie.Value, CookieType.PINsafe);
                if (newCookieValue != null) {
                    string cookieDomain = pinsafeCookie.Domain;
                    pinsafeCookie = new HttpCookie(COOKIE_NAME, newCookieValue) {Path = "/", Domain = cookieDomain};
                    context.Response.Cookies.Add(pinsafeCookie);
                }
//                Debug.WriteLine(LOG_SOURCE, "User already authenticated to Swivel");
            }
        }

        public void Dispose() {
            // Nothing to dispose.
        }

        private bool CheckAuthenticationRequired(string requestUrl) {
            requestUrl = requestUrl.ToLower();
            if (settings.ExcludedPaths.Any(path => !CheckPath(requestUrl, path))) {
                return false;
            }
            bool required = CheckPath(requestUrl, settings.LogonUrl);
            if (!required) {
                return false;
            }
            required = CheckPath(requestUrl, settings.ChangePinUrl);
            if (!required) {
                return false;
            }
            required = CheckPath(requestUrl, settings.ChangePasswordUrl);
            if (!required) {
                return false;
            }
            required = CheckPath(requestUrl, settings.AuthUrl);
            return required;
        }

        private static bool CheckPath(string requestUrl, string path) {
            string lpath = path.ToLower();
            if (lpath.StartsWith("./")) {
                // Relative path: compare backwards.
                string[] requestParts = requestUrl.Split('/');
                string[] pathParts = lpath.Split('/');
                for (int rdx = requestParts.Length - 1, pdx = pathParts.Length - 1; rdx >= 0 && pdx >= 0; rdx--, pdx--) {
                    if (pdx == 0) {
                        return false; // We've matched up to the relative bit.
                    }
                    if (!requestParts[rdx].Equals(pathParts[pdx])) {
                        break;
                    }
                }
            } else if (requestUrl.StartsWith(lpath)) {
                return false;
            }
            return true;
        }
    }
}
