﻿using System;
using System.Text;
using System.Web.UI;
using com.swivelsecure.pinsafe;

namespace Web._forms {
	public partial class Reset : Page {

		private const string LOG_SOURCE = "RESET";

		private readonly TextSettings text;
		private readonly PINsafeSettings settings;
		private readonly DebugLogger logger;
		private bool requested;
		private readonly PINsafeClient client;

		public Reset() {
			text = new TextSettings();
			settings = PINsafeSettings.SettingsFromConfiguration;
			logger = settings.Logger;
			client = new PINsafeClient(settings, text);
			Load += PageLoad;
		}

		private string Username {
			get { return txtUsername.Text; }
			set { txtUsername.Text = value; }
		}

		private string ResetCode {
			get { return txtResetCode.Text; }
		}

		protected void PageLoad(object sender, EventArgs e) {
			if (!settings.ResetPinEnabled) {
				trUsername.Visible = false;
				trResetCode.Visible = false;
				trResetPwd.Visible = false;
				tdButtons.Visible = false;
				SetError(text.ResetDisabled);
				return;
			}
			if (!settings.ResetPasswordEnabled) {
				trResetPwd.Visible = false;
			}
			if (!IsPostBack) {
				if (!String.IsNullOrEmpty(Request["Source"])) {
					hfRedirectUrl.Value = Request["Source"];
				}
				if (!String.IsNullOrEmpty(Request["username"])) {
					Username = Request["username"];
				}
			}
			CheckResetCode();
			lblUsername.Text = text.Username + ":";
			if (trResetCode.Visible) {
				lblResetCode.Text = text.ResetCode + ":";
			}
			imgLogo.Src = text.LogoFile;
			Title = text.ResetPageTitle;
			lblTitle.Text = text.ResetPageHeader;
			lblFooter.Text = text.ResetPageFooter;
		}

		private void CheckResetCode() {
			if (requested) {
				txtUsername.ReadOnly = true;
				btnReset.Visible = true;
				btnReset.Text = text.ResetButton;
				btnGetCode.Visible = false;
				trResetCode.Visible = true;
			} else {
				txtUsername.ReadOnly = false;
				btnReset.Visible = false;
				btnGetCode.Visible = true;
				btnGetCode.Text = text.GetResetCodeButton;
				trResetCode.Visible = false;
			}
		}

		protected void GetCodeClick(object sender, EventArgs e) {
			if (client.RequestResetCode(Username)) {
				requested = true;
				logger.WriteLine(LOG_SOURCE, "Reset code requested for user \"{0}\"", Username);
				CheckResetCode();
			} else {
				SetError("Request reset code failed");
				logger.WriteLine(LOG_SOURCE, "Reset code request failed for user \"{0}\": {1}", Username, client.ErrorMessage);
			}
		}

		private void SetError(string message) {
			SetComment(message, Login.CSS_ERROR_CLASS);
		}

		private void SetInfo(string message) {
			SetComment(message, Login.CSS_INFO_CLASS);
		}

		private void SetComment(string message, string cssClass) {
			lblResult.Text = message;
			lblResult.CssClass = cssClass;
		}

		protected void ResetClick(object sender, EventArgs e) {
			if (!settings.ResetPinEnabled) {
				SetError(text.ResetDisabled);
				return;
			}
			if (client.ResetPIN(Username, ResetCode)) {
				logger.WriteLine(LOG_SOURCE, text.ResetPinSuccess, Username);
				StringBuilder linkText = new StringBuilder(text.ResetPinSuccess);
				if (settings.ResetPasswordEnabled && chkResetPwd.Checked) {
					// Reset the password as well.
					ActiveDirectoryClient adClient = new ActiveDirectoryClient(settings, text);
					string newPassword = adClient.ResetPassword(Username);
					if (newPassword == null) {
						SetError(String.Format(text.ResetPasswordFailed, Username, String.IsNullOrEmpty(adClient.Error) ? "" : adClient.Error));
					} else {
						SetInfo(String.Format(text.ResetPasswordSuccess, Username));
						logger.WriteLine(LOG_SOURCE, String.Format(text.ResetPasswordSuccess, Username)); // + String.Format(" New Password is \"{0}\"", newPassword));
						// Send a message with the new password.
						if (!adClient.SendResetPasswordAlert(Username, newPassword)) {
							SetError("Password was reset, but password was not sent to user");
						}
					}
				}
				string source = hfRedirectUrl.Value;
				if (String.IsNullOrEmpty(source)) {
					source = Request["Source"];
				}
				if (!String.IsNullOrEmpty(source)) {
					linkText.Append("<br/>Click <a href='{0}'>here</a> to continue");
				}
				SetInfo(string.Format(linkText.ToString(), Username, source));
			} else {
				logger.WriteLine(LOG_SOURCE, text.ResetPinFailed, Username, client.ErrorMessage);
				SetError(String.Format(text.ResetPinFailed, Username, client.ErrorMessage));
			}
		}
	}
}